// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					  (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Ccf.cc	--- interface between the CAL and the CIF in the CCF
//
//	DESCRIPTION:
//		This defines the interface between the CAL and the CIF in the CCF
//
//	CHANGE HISTORY:
//		UL, 1997-10-20: 1.0.0: - skeleton
//		UL, 1998-??-??: 1.1.0: - working version
//		UL, 1999-02-10: 1.1.1: - infer scope of CIF constituent from contents
//								 of "SCOPE" attribute (was: "INSTRUME")
//		UL, 1999-05-12: 1.1.2: - changed fatal -> error and changed all
//								 error/warning tags
//		UL, 1999-10-11: 1.1.3: - no longer check value of CATEGORY keyword
//								 against "XMMCCF"; print value as message
//								 if SAS_VERBOSITY >= VERBOSE
//		UL, 2000-02-03: 1.2.0: - make presence of EXTSEQU/EXTSEQID in CCF
//								 override constituents optional; if they
//								 are not found: EXTSEQU = 'ANY' and
//								 EXTSEQUID = <list of block names>
//		UL, 2000-04-27: 1.3.0: - be robust against VALDATE/DATE possibly
//								 containing invalid dates
//		UL, 2000-05-23: 1.4.0: - SUBDATE is now optional
//		UL, 2000-10-03: 1.5.0: - various changes related to support for
//								 CCF constituent search path (SAS_CCFPATH)
//		UL, 2000-12-18: 1.6.0: - added class CcfConstituentBlock
//		UL, 2001-03-12: 1.7.0: - fixed SPR 1949: Wrong verbosity level
//		UL, 2001-03-15: 1.8.0: - changed error->fatal to signal that
//								 further execution does not make sense
//		UL, 2001-09-27: 1.9.0: - bug fix: if a CCF constituent was specified
//								 with path component (in CIF or as replacement
//								 on command line) it was nevertheless searched
//								 along SAS_CCFPATH
//
//	AUTHOR:
//		U. Lammers, Tue Oct 14 14:45:35 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Indexing of the XMM Current Calibration File, U. Lammers,
//			 15/07/97
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/Ccf.cc,v 1.7 2009/06/23 13:46:02 aibarra Exp $
//
#include <fstream>
#include <iostream>
#include <string>
#include <sys/types.h>
#include <dirent.h>
#include <algorithm>
#include <functional>

#include "errstr.h"
#include "common.h"
#include "stringutil.h"
#include "CalOalUtility.h"
#include "StringIntMap.h"
#include "Ccf.h"
#include "Dal.h"
#include "FileLocator.h"
#include "SasFileLocator.h"

using namespace std;

//
//	CIF entries
//

//
//	assignment op/copy constructor
//
CifEntry &CifEntry::operator=(const CifEntry &entry)
{
	if (this == &entry)
		return *this;

	_instrument = entry._instrument;
	_calDataType = entry._calDataType;
	_issue = entry._issue;
	_fileName = entry._fileName;
	_validityDate = entry._validityDate;
	_submissionDate = entry._submissionDate;
	_extensionKey = entry._extensionKey;
	_extensionSequence = entry._extensionSequence;

	return *this;
} // CifEntry::operator=()


CifEntry::CifEntry(const CifEntry &entry) : CalChild<Ccf>(entry.parent())
{
	operator=(entry);
} // CifEntry::CifEntry()

//
//	construct from calibration file name
//
CifEntry::CifEntry(const string &calibrationFileName, const Ccf *ccf)
	: CalChild<Ccf>(ccf)
{
	//	open the data set
	_fileName			= calibrationFileName;
	DataSet *set = open(DataSetServer::Low, false);

	errstr<<"CATEGORY of CCF constituent `"<<set->qualifiedName()<<"': "
		  <<set->attribute("CATEGORY")->asString()
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	_instrument			= Xmm::toInstrument(set->attribute("SCOPE")->
							asString());
	_calDataType		= Ccf::toTypeId(set->attribute("TYPEID")->
							asString());
	_issue				= set->attribute("ISSUE")->asInt();
	string validityDate = set->attribute("VALDATE")->asString();
	if (STime::isValidTime(validityDate))
		_validityDate		= STime(validityDate);

	if (set->hasAttribute("SUBDATE")) {
		string submissionDate(set->attribute("SUBDATE")->asString());
		if (STime::isValidTime(submissionDate))
			_submissionDate	= STime(submissionDate);
	} else
		errstr<<"CCF constituent `"<<set->qualifiedName()<<"' has "
				"no SUBDATE"<<Sas::warning("NoSUBDATEKeyword");

	if (set->hasAttribute("EXTSEQU") && set->hasAttribute("EXTSEQID")) {
		_extensionKey		= set->attribute("EXTSEQU")->asString();
		_extensionSequence	= set->attribute("EXTSEQID")->asString();
	} else {
		errstr<<"CCF constituent `"<<set->qualifiedName()<<"' does not "
				"contain EXTSEQU/EXTSEQID keywords"
			  <<Sas::warning("MissingKeywords");
		_extensionKey = "ANY";
		unsigned N = set->blocks();
		for (unsigned n=0; n<N-1; ++n)
			_extensionSequence += set->block(n)->name() + ' ';
		_extensionSequence += set->block(N-1)->name();
	}

	//	close data set
	dataSetServer->close(set);
} // CifEntry::CifEntry


bool CifEntry::hasBlock(const string &truncatedBlockName) const
{
	StringIntMap<int> _map(_extensionSequence);
	return _map[truncatedBlockName] >= 0;
} // CifEntry::hasBlock


//	query extension number
string CifEntry::queryBlockName(const string &truncatedBlockName,
								unsigned *blockNr) const
{
	string calBlockName;
	StringIntMap<int> _map(_extensionSequence);
	int calBlockId = _map[truncatedBlockName];

	if (calBlockId >= 0)
		//
		//	this extension exists
		//
		if (truncatedBlockName.find("X-", 0, 2) != string::npos ||
			_extensionKey == "ANY")
			//
			//	name starts with "X-" or the key is "ANY", i.e.,
			//
			calBlockName = truncatedBlockName;
		else
			//
			//	put the extension sequence key in front
			//
			calBlockName = _extensionKey + "-" + truncatedBlockName;
	else
		errstr<<"unknown block `"<<truncatedBlockName
			  <<"'"<<Sas::fatal("UnknownBlock");

	if (blockNr)
		*blockNr = calBlockId;

	return calBlockName;
} // CifEntry::queryBlockName()


bool operator<(const CifEntry &e1, const CifEntry &e2)
{
	return Xmm::toString(e1.instrument())+':'+Ccf::toString(e1.calDataType()) <
		   Xmm::toString(e2.instrument())+':'+Ccf::toString(e2.calDataType());
} // operator<


//
//	open
//
DataSet *CifEntry::open(DataSetServer::MemoryModel memModel, bool logAccess)
const
{
	if (!parent())
		errstr<<"CifEntries can only be opened after they have been "
				"registered with a CIF"
			  <<Sas::fatal("CifEntryNotRegisteredWithCif");

	string ccfPath = parent()->ccfSearchPath();
	if (ccfPath.empty()) {
		char *ccfPathPtr = getenv(Ccf::SAS_CCFPATH);
		if (ccfPathPtr)
			ccfPath = string(ccfPathPtr);
		if (!ccfPath.empty())
			ccfPath += ':';

		ccfPath += parent()->cifDirName();
		ccfPath += ":.";
	}
	// GV: fileLocator is the name of an extern in SasFileLocator.h, so here
	// we call it myfileLocator.
	FileLocator myfileLocator(ccfPath);

	_fileName = fileName();
	if (!_fileName.empty() && _fileName[0] != '/' && _fileName[0] != '.')
		//	just file name - look along SAS_CCFPATH for location
		_fileName = myfileLocator.locate(_fileName);
	DataSet *set = dataSetServer->open(_fileName, Dal::Read, memModel);

	if (logAccess)
		CcfAccessLog::instance().insert(*this);

	return set;
} // CifEntry::open()


//	destructor
CifEntry::~CifEntry() {}


// print CifEntry to a stream
ostream &operator<<(ostream &os, const CifEntry &e)
{
	os<<"CifEntry{"<<Xmm::toString(e._instrument)<<", "
	  <<Ccf::toString(e._calDataType)<<", "
	  <<e._issue<<", "<<e._fileName<<", "
	  <<e._validityDate<<"}";

	return os;
} // operator<<


//
//	CCF class implementation
//
static string __tstr[] = { TYPE_STR, "" };
StringIntMap<Ccf::CalDataTypes> Ccf::_calDataTypes(__tstr);

const char *const Ccf::SAS_CCF				= "SAS_CCF";
const char *const Ccf::SAS_DEFAULT_CIFNAME	= "ccf.cif";
const char *const Ccf::SAS_CCFPATH			= "SAS_CCFPATH";
const char *const Ccf::SAS_CCFFILES			= "SAS_CCFFILES";

Ccf::Ccf(const string cifFileName)
{
	string cifName = cifFileName;
	if (cifName.empty()) {
		//	no file name has been specified - inquire environment
		//	for location of CIF to use
		char *cifPtr;
		if ((cifPtr = getenv(Ccf::SAS_CCF)))
			cifName = string(cifPtr);
		else
			errstr<<"No file name for Calibration Index File specified and "
					"environment variable `"
				  <<Ccf::SAS_CCF<<"' is not set either - "
					"cannot continue."
				  <<Sas::fatal("CannotLocateCIF");
	}

	_cifDirName = "./";
	DIR *dir = opendir(cifName.c_str());
	if (dir) {
		closedir(dir);
		_cifDirName = cifName + '/';
		cifName = _cifDirName + Ccf::SAS_DEFAULT_CIFNAME;
	} else {
		ifstream f(cifName.c_str());
		if (!f)
			errstr<<"`"<<cifName<<"' is neither the name of a directory nor "
					"the name of a CIF"
				  <<Sas::fatal("NoCifSpecified");
		else {
			string::size_type pos =  cifName.rfind('/');
			if (pos != string::npos)
				_cifDirName = cifName.substr(0, pos);
		}
	}
	_cifFileName = cifName;
} // Ccf::Ccf()


Ccf::CalDataTypes Ccf::toTypeId(const string &typeString)
{
	return _calDataTypes[typeString];
} // Ccf::toTypeId


const string &Ccf::toString(CalDataTypes dataTypeId)
{
	return _calDataTypes[dataTypeId];
} // Ccf::toString


const string &Ccf::toString(unsigned dataTypeId)
{
	return toString(static_cast<CalDataTypes>(dataTypeId));
} // Ccf::toString


Ccf::~Ccf()
{
}


void Ccf::print(ostream &os) const
{
	os<<"Ccf{"<<_cifFileName<<"}";
} // Ccf::print()


// print CCF to a stream
ostream &operator<<(ostream &os, const Ccf *ccf)
{
	ccf->print(os);
	return os;
} // operator<<



ostream &operator<<(ostream &os, const CcfAccessLog &log)
{
	for (set<CifEntry>::const_iterator i = log.begin(); i != log.end(); ++i)
		os<<*i<<endl;
	// Issues warnings about possible expired CCF constituents
	log.evaluate();

	return os;
} // operator<<


#define CODE																\
	_algoid = 0;															\
	if (_block->hasAttribute("ALGOID"))										\
		_algoid = _block->attribute("ALGOID")->asInt();						\
																			\
	errstr<<"Accessing table `"<<_block->qualifiedName()<<"' with ALGOID="	\
		  <<_algoid<<Sas::message(Msg::Lib, Msg::Verbose);					\
																			\
	if (_algoid > acceptAlgoIdsLessEqThan)									\
		errstr<<"The table/array `"<<_block->qualifiedName()				\
			  <<"' specifies ALGOID="<<_algoid<<" but this version of the"	\
			  <<" CAL merely supports 0<=ALGOID<="<<acceptAlgoIdsLessEqThan	\
			  <<" - you may want to upgrade to a more recent version of the"\
				" CAL that knows about the reported ALGOID."				\
			  <<Sas::fatal("UnsupportedAlgoId");
/* AI gcc4 porting
 * Statement template<> added.
 */
template<>
CcfConstituentBlock<Table>::CcfConstituentBlock(const DataSet *set,
											const string &name,
											unsigned acceptAlgoIdsLessEqThan,
											bool treatBlockAsOptional)
	: _algoid(0), _block(0)
{
	if (set->hasBlock(name) || !treatBlockAsOptional) {
		//	access the table - if it is missing and not optional the following
		//	will cause an error
		_block = set->table(name);
		CODE
	}
} // CcfConstituentBlock<Table>::CcfConstituentBlock()

template<>
CcfConstituentBlock<Table>::CcfConstituentBlock(const DataSet *set,
											unsigned nr,
											unsigned acceptAlgoIdsLessEqThan,
											bool treatBlockAsOptional)
{
	if (nr < set->blocks() || !treatBlockAsOptional) {
		//	access the table - if it is missing and not optional the following
		//	will cause an error
		_block = set->table(nr);
		CODE
	}
} // CcfConstituentBlock<Table>::CcfConstituentBlock()

template<>
CcfConstituentBlock<Array>::CcfConstituentBlock(const DataSet *set,
											const string &name,
											unsigned acceptAlgoIdsLessEqThan,
											bool treatBlockAsOptional)
{
	if (set->hasBlock(name) || !treatBlockAsOptional) {
		//	access the array - if it is missing and not optional the following
		//	will cause an error
		_block = set->array(name);
		CODE
	}
} // CcfConstituentBlock<Array>::CcfConstituentBlock()

template<>
CcfConstituentBlock<Array>::CcfConstituentBlock(const DataSet *set,
											unsigned nr,
											unsigned acceptAlgoIdsLessEqThan,
											bool treatBlockAsOptional)
{
	if (nr < set->blocks() || !treatBlockAsOptional) {
		//	access the array - if it is missing and not optional the following
		//	will cause an error
		_block = set->array(nr);
		CODE
	}
} // CcfConstituentBlock<Array>::CcfConstituentBlock()

#undef CODE

//	CcfAccessLog
CcfAccessLog CcfAccessLog::_instance;


CcfAccessLog::CcfAccessLog() : std::set<CifEntry>()
{
	// FIXME
	// GV: I would like to issue messages here, but I cannot because this
	// constructor is part of a chain of global objects, and there is no
	// guarantee that the error stream is initialized at this time.
	std::string file("config/ccf.conf");
	// fileLocator comes from SasFileLocator.h
	std::string ccfcnffile = fileLocator.locate(file, "", false, false);
	if(ccfcnffile != "") {
		//	errstr << "Did not find any CCF configuration file. All CCF constituents are considered valid."
		//  			 << Sas::message(Msg::Lib, Msg::Silent);
		//	} else {
		//  	  errstr << "Found the CCF configuration file " + ccfcnffile + "."
		//  			 << Sas::message(Msg::Lib, Msg::Silent);
		ifstream is(ccfcnffile.c_str());
		// lines having '#' in the first column are considered comments
		// For the other lines the format is as follows:
		// <constituent name> [optional description]
		// XMM_MISCDATA_1234 Wrong for all SAS versions after 5.3
		std::string line;
		while(getline(is,line) && is.good()){
		if (line[0] == '#')
			continue;
		if(line.find_first_not_of(' ') == std::string::npos)
			continue;
		std::string::size_type p = line.find_first_of(' ');
		if (p == std::string::npos)
			invalidCcf_.insert(std::make_pair(line.substr(0, p),
							  std::string("")));
		else
			invalidCcf_.insert(std::make_pair(line.substr(0, p),
							   line.substr(p, std::string::npos)));
		}
		is.close();
		//	  for(std::map<std::string, std::string>::const_iterator K =
		//		invalidCcf_.begin(); K != invalidCcf_.end(); ++K)
		//		std::cout << (*K).first << ": " << (*K).second << std::endl;
		//  		errstr << (*K).first << " : " << (*K).second
		//   << Sas::message(Msg::Lib, Msg::Silent);
	}
	//}
}


void CcfAccessLog::evaluate(const CifEntry & c) const
{
	Filename::StringPair sp = Filename::split(c.fileName());
	std::map<std::string, std::string>::const_iterator p =
											invalidCcf_.find(sp.second);
	if (p != invalidCcf_.end()) {
		errstr<<"The CCF constituent "<<c.fileName()
			  <<" has been accessed, but it is on the list of expired CCF "
				"constituents. This is the associated message: "
			  <<std::endl<<(*p).first<<" : "<<(*p).second<<std::endl
			  <<"Please consult the CCF release notes at the XMM-Newton SOC."
			  <<Sas::warning("ExpiredCcfConstituent");
	}
} // CcfAccessLog::evaluate()


void CcfAccessLog::evaluate() const
{
	for (const_iterator p = begin(); p != end(); ++p)
		evaluate(*p);
} // CcfAccessLog::evaluate()


std::pair<std::set<CifEntry>::iterator, bool> CcfAccessLog::insert(
															const CifEntry & c)
{
//  	errstr << "CcfAccessLog::insert() " << c.fileName()
//  		   << Sas::message(Msg::Lib, Msg::Noisy);
	evaluate(c);
	return std::set<CifEntry>::insert(c);
} // CcfAccessLog::insert()


void CcfAccessLog::reset()
{
	erase(begin(), end());
}


CcfAccessLog::~CcfAccessLog() {}

CcfAccessLog &CcfAccessLog::instance() { return _instance; }


