// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997 - 2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	KeplerOrbitSimulator.h --- simulates objects in Kepler orbit around a body
//
//	DESCRIPTION:
//		This is a simple hack to mimic the behavior of a rigid small body which
//		moves in a Keplerian orbit around a massive second body.
//
//	CHANGE HISTORY:
//		UL, 1999-06-11: 1.0.0: - first version
//		UL, 2001-01-05: 1.1.0: - added eccentricity to orbital elements
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/KeplerOrbitSimulator.h,v 1.5 2004/08/04 12:18:14 rsaxton Exp $
//

#ifndef ORBITSIMULATOR_H
#define ORBITSUMULATOR_H

#include "Celestial.h"

static const double EARTH_RADIUS = 6378.137;		//	[km]

struct OrbitalElements {
	double	period,				//	s
			apogeeHeight,		//	km
			perigeeHeight,		//	km
			eccentricity,		//	0<e<1
			inclincation,		//	orbit inclination [rad]
			ascendingNode,		//	R.A. of ascending node [rad]
			perigeeArgument;	//	argument of perigee [rad]
};


class KeplerOrbitSimulator {
public:
	KeplerOrbitSimulator() {}
	KeplerOrbitSimulator(const OrbitalElements &orbitalElements,
						 double initalPerigeePassTime = 0.);
		//
		//	PURPOSE:
		//		Construct a simulator object which can calculate the
		//		position and velocity of an object which moves along
		//		a Keplerian trajectory around a massive body with
		//		constant angular velocity
		//
		//	ARGUMENTS:
		//		orbitalElements		: structure with quantities defining
		//							  the Kepler orbit
		//		initialPerigeePassTime: point in time defining the inital
		//							  pass of the object through the
		//							  perigee such that
		//								KeplerOrbitSimulator::position(
		//									initalPerigeePass +
		//										n * orbitalElements.T)
		//							  gives the position of the perigee
		//


	virtual CartesianVector position(double time) const;
		//
		//	PURPOSE:
		//		Return the position of the object as a Cartesian vector
		//		in the Earth-centered equatorial J2000 reference system
		//		(units km) at a given point in time
		//
		//	ARGUMENTS:
		//		time	: time for which position is requested [s after
		//				  initial perigee pass time]
		//


	virtual CartesianVector velocity(double time) const;
		//
		//	PURPOSE:
		//		Return the velocity of the object as a Cartesian vector
		//		in the Earth-centered equatorial J2000 reference system
		//		(units km/s) at a given point in time
		//
		//	ARGUMENTS:
		//		time	: time for which position is requested [s after
		//				  initial perigee pass time]
		//


	const OrbitalElements& orbitalElements() const { return _orbitalElements; }
		//
		//	PURPOSE:
		//		Return orbital elements this simulator is valid for
		//


	virtual ~KeplerOrbitSimulator() { }
		//
		//	PURPOSE:
		//		Destroy KeplerOrbitSimulator object constructed with
		//		above constructor
		//

protected:
	void orbitalElements(const OrbitalElements &orbitalElements,
						 double initalPerigeePassTime = 0);
		//
		//	PURPOSE:
		//		Define orbital elements and time of inital perigee pass
		//		for this object
		//
		//	ARGUMENTS:
		//		orbitalElements	: Keplerian orbital elements for this
		//						  simulator
		//		initialPerigeePassTime:
		//						  a scalar time in [s] which defines the
		//						  inital perigee pass of the object and thus
		//						  defines the phase of the object in the
		//						  orbit; calling position()/velocity() for
		//						  times (n>=0):
		//							initalPerigeePassTime + n*P
		//						  shall yield the position and velocity
		//						  at the point of perigee passage
		//

private:
	OrbitalElements	_orbitalElements;		//	orbital elements
	double			_initialPerigeePassTime;//	first perigee passage
	AttitudeMatrix	_a313;					//	rotation matrix
	double			_phi0,					//	phase 0
					_omega,					//	angular velocity
					_a,						//	semi-major axis
					_b,						//	semi-minor axis
					_c;						//	difference center - focus
											//	ellipse
	bool _orbitalElementsSet;
	
	//
	//	methods
	//
	void init();
};

#endif
