// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OdfDesignators.h --- Observation Data File designators related classes
//
//	DESCRIPTION:
//		Various classes are defined in conjunction with needed functionality
//		to access information in an ODF.
//
//	CHANGE HISTORY:
//		UL, 2001-12-06: 1.0.0: - initial version; partially extracted from
//								 Odf.h
//
//	AUTHOR:
//		U. Lammers, Thu Dec  6 16:05:50 MET 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Observation Data File Interface Control Document, K. Galloway
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/OdfDesignators.h,v 1.5 2009/07/06 15:04:45 aibarra Exp $
///

#ifndef ODF_DESIGNATORS_H
#define ODF_DESIGNATORS_H

#include <string>
#include <iosfwd>

#include "Xmm.h"

//
//	an OdfSet represents a data set with a name compliant with [2];
//	instantiation from a name; access methods to access the various
//	fields
//
class OdfSet {
public:
	OdfSet();
	OdfSet(const std::string &name);

	enum Type { TNONE = -1, EVENT_LIST, IMAGE, AUXILIARY, HOUSKEEPING,
				SPACECRAFT, SUMMARY };
	enum Format { FNONE = -1, COMPRESSED_FITS, COMPRESSED_ASCII, FITS, ASCII };

	const std::string		name() const { return _dirname + '/' + _basename; }
	const std::string		&dirname() const { return _dirname; }
	const std::string		&basename() const { return _basename; }
	unsigned			revolutionNr() const { return _revolutionNr; }
	const std::string		&proposalId() const;
	Xmm::Instrument		instrument() const { return _instr; }
	bool				scheduleFlagApplicable() const {
							return _scheduleFlagApplicable; }
	bool				scheduled() const { return _scheduleFlagApplicable &&
												   _scheduled; }
	bool				unscheduled() const { return _scheduleFlagApplicable &&
												   !_scheduled; }
	unsigned			exposureNr() const { return _exposureNr; }
	unsigned		    ccdNr() const { return _ccdNr; }
	unsigned			nodeNr() const { return _nodeNr; }
	Xmm::DataMode		dataMode() const { return _dataMode; }
	OdfSet::Type		type() const { return _type; }
	OdfSet::Format		format() const { return _format; }

	static bool			isValid(const std::string &name);
	bool				isValid() const { return _isValid; }

	~OdfSet() {}

private:
	//	data members
	bool				_isValid;
	std::string				_basename,
						_dirname;
	unsigned			_revolutionNr;
	std::string				_proposalId;
	Xmm::Instrument		_instr;
	bool				_scheduleFlagApplicable, _scheduled;
	unsigned			_exposureNr,
						_ccdNr,
						_nodeNr;
	Xmm::DataMode		_dataMode;
	OdfSet::Type		_type;
	OdfSet::Format		_format;
};

std::ostream &operator<<(std::ostream &os, const OdfSet &fname);

//
//	an identifier which uniquely identifies an instrument + exposure
//	in a given ODF; construction from 6-character string e.g. "R1S001";
//	methods to access the three fields
//
class InstrumentExposureIdentifier {
public:
	InstrumentExposureIdentifier(const std::string &id);
		//
		//	PURPOSE:
		//		Instantiate from 6-char string ("R1S001") or ODF ICD
		//		compliant set name
		//

	Xmm::Instrument		instrument() const { return _instr; }
	bool				scheduleFlagApplicable() const {
							return _scheduleFlagApplicable; }
	bool				scheduled() const { return _scheduleFlagApplicable &&
											_scheduled; }
	bool				unscheduled() const { return _scheduleFlagApplicable &&
											!_scheduled; }
	unsigned			exposureNumber() const { return _expNr; }

	std::string			exposureId() const;
	std::string			instrumentExposureId() const;

private:
	Xmm::Instrument		_instr;
	bool				_scheduleFlagApplicable, _scheduled;
	unsigned			_expNr;
};

#endif
