// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Xmm.c	--- implementation of class XMM
//
//	CHANGE HISTORY:
//		UL, 1997-??-??: 1.0.0: - first version
//		UL, 1998-08-27: 1.1.0: - changed PN CCD numbering scheme according
//								 to quadrantId * 3 + ccdId + 1
//		UL, 1999-02-23: 1.2.0: - new global object of type XmmPayLoad
//		UL, 1999-04-01: 1.3.0: - in-line with Xmm.h 2.3.0
//		UL, 1999-04-14: 1.4.0: - small change in mode detection logic for
//								 EMOS
//		UL, 1999-04-23: 1.4.1: - change in logic to detect EMOS mode W6
//								 (suggested by Jean Ballet)
//		UL, 1999-05-12: 1.4.2: - changed fatal -> error and changed all
//								 error/warning tags
//		UL, 1999-06-11: 1.4.3: - changed launch date to 1999-12-15T00:00:00
//							   - added orbital elements
//		UL, 1999-06-15: 1.4.4: - added implementation o
//								 EPNcam::toQuadrantColumnNumber()
//		UL, 1999-06-18: 1.4.5: - added another version of
//								 EPNcam::toQuadrantColumnNumber()
//		UL, 1999-06-24: 1.4.6: - added initialization routine for global
//								 xmmPayLoad object
//		UL, 1999-10-11: 1.5.0: - updated value of Xmm::missionStartTime
//		UL, 1999-10-20: 1.6.0: - added Xmm::emptyDataModeVec
//							   - added XmmInstrument::periodicHKmodes()
//		UL, 1999-11-17: 1.7.0: - brought logic in RGScam::toMode() in
//								 line with v7.0 [1] (changed SCITYPES)
//		UL, 2000-04-04: 1.8.0: - brought logic in RGScam::toMode() in
//								 line with v9.0 of [1] (changed SCITYPES)
//		UL, 2000-04-10: 1.9.0: - bug fix in EPNcam::toQuadrantColumnNumber()
//							   - brought in line with issue 2.4 of [2]
//		UL, 2000-05-24: 1.10.0:- EMOScam::toMode()/EPNcam::toMode() need
//								 now vector of CoreModeParameters as argument
//		UL, 2000-09-21: 1.11.0:- toMode() now returns pair<Mode, bool>
//		UL, 2000-09-29: 1.12.0:- toIntergralId() gives list of allowed
//								 values in error message when passed illegal
//								 string
//							   - various transparent changes to bring in
//								 line with changes in Xmm.h
//		UL, 2000-10-17: 1.13.0:- added XmmInstrument::scienceDataModes()
//		UL, 2000-11-07: 1.14.0:- change in mode detection logic for EMOS:
//								 outer CCDs are always operated in
//								 FullFrameMode mode; the mode detection logic
//								 cannot determine this from the mere contents
//								 of WINDOWX0/Y0/DX/DY + FRAMETIME alone
//								 because if the inner CCD is in RFS the
//								 frame time of the outer ones are extended
//								 to be an integer multiple of the inner
//								 FRAMETIME; this leads to ambiguities.
//		UL, 2001-01-10: 1.15.0:- added complete NASA/OIG TLE history covering
//								 period [1999-12-10, 2001-01-10]
//		UL, 2001-02-28: 2.17.0:- added couple of new classes:
//									* XmmGroundStation
//									* XmmMission
//		UL, 2001-03-22: 2.18.0:- added
//									* EMOScam::toHBR(Ccd)
//		UL, 2001-07-13: 2.19.0:- tweaked EPN mode detection logic such that
//								 ExtendedFullFrame is also recognized if
//								 FRAMETIME has a wrong value!
//		UL, 2001-07-16: 2.19.1:- associate DiscardedLinesData with Diagnostic
//								 mode
//							   - added ScienceEngineeringDataModes
//		UL, 2001-10-29: 2.20.0:- added 'using namespace std'
//		UL, 2002-03-07: 2.21.0:- EPNcam::toMode() now checks for F1294
//								 attribute to distinguish between FF and eFF
//								 (see also ODF-ICD DCP-140)
//		UL, 2002-04-16: 2.22.0:- updated GS propagation delays with latest
//								 data from Ed Serpell (should not have
//								 considerable impact on accuracy of time tags)
//		UL, 2003-03-12: 2.23.0:- extended mode-detection logic for MOS+OM
//								 according to email GV/MG 2003-03-07
//		UL, 2003-05-23: 2.24.0:- updated orbital TLE elements to 2003-04-28 and
//								 added mean anomaly
//
//	REFERENCES:
//		[1]: RGS Telemetry Structure, RGS-MSSL-SW-002, v9.0
//		[2]: ICD: Observation and Slew Data Files, K. Galloway
//
//  AUTHOR:
//		U. Lammers, Thu May 28 17:44:55 MET DST 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/Xmm.cc,v 1.49 2010/11/16 13:52:51 aibarra Exp $
//

#include <algorithm>

#include "slatec++.h"
#include "Tracer.h"
#include "Xmm.h"
#include "Dal.h"
#include "DalUtils.h"
#include "StringPool.h"
#include "stringutil.h"

using namespace std;

static string _undef("UNDEFINED");

//
//	auxiliary function
//
namespace {

int toIntegralId(const StringPool &pool, const string &keyString,
				 const string &desc)
{
	if (keyString == _undef)
		return -1;

	string idstring(pool.expand(keyString));
	std::vector<string>::const_iterator sp = pool.begin();
	while (sp != pool.end()) {
		if ((*sp) == idstring)
		{
			break;
		}
		++sp;
	}

	if (sp != pool.end())
		return sp - pool.begin();
	else {
		errstr<<"`"<<keyString<<"' is not a valid "<<desc<< " identifier - "
				"the list of valid identifiers is:\n";
		unsigned N = pool.size();
		for(unsigned n=0; n<N; ++n)
			errstr<<pool[n]<<endl;
		errstr<<Sas::error("IllegalIdentifier");
	}

	return 0;
} // toIntegralId()


string dataType(const vector<const Attributable *> &av) {
	string dataType;
	try {
		dataType = uppercase(DalUtils::attribute(av, "DATATYPE")->asString());
	} catch (...) {
		errstr<<"Unable to find mandatory attribute `DATATYPE' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributeNotFound");
	}

	return dataType;
} // dataType()


vector<int> windowCoordinates(const vector<const Attributable *> &av)
{
	vector<int> v;
	try {
		v.push_back(DalUtils::attribute(av, "WINDOWX0")->asInt());
		v.push_back(DalUtils::attribute(av, "WINDOWY0")->asInt());
		v.push_back(DalUtils::attribute(av, "WINDOWDX")->asInt());
		v.push_back(DalUtils::attribute(av, "WINDOWDY")->asInt());
	} catch(...) {
		errstr<<"Unable to find manatory attributes `WINDOWX0', `WINDOWY0', "
				"`WINDOWDX', `WINDOWDY' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return v;
} // windowCoordinates()

pair<unsigned, string> frameTime(const vector<const Attributable *> &av)
{
	pair<unsigned, string> ft;
	try {
		const Attribute *frt = DalUtils::attribute(av, "FRMTIME");
		ft = make_pair(unsigned(frt->asInt()), frt->units());
	} catch(...) {
		errstr<<"Unable to find manatory attribute `FRMTIME' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return ft;
} // frameTime()

unsigned ccdMode(const vector<const Attributable *> &av)
{
	unsigned ccdm = 0;
	try {
		ccdm = DalUtils::attribute(av, "CCDMODE")->asInt();
	} catch(...) {
		errstr<<"Unable to find manatory attribute `CCDMODE' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return ccdm;
} // ccdMode()

pair<unsigned, unsigned> emdh(const vector<const Attributable *> &av)
{
	pair<unsigned, unsigned> emdhval = make_pair(0, 4095);

	try {
		emdhval = make_pair(DalUtils::attribute(av, "EMDHLOW")->asInt(),
							DalUtils::attribute(av, "EMDHUPP")->asInt());
	} catch(...) {
		errstr<<"Unable to find manatory attributes `EMDHLOW', `EMDHUPP' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return emdhval;
} // emdh()

unsigned sciType(const vector<const Attributable *> &av)
{
	unsigned st = 0;
	try {
		st = DalUtils::attribute(av, "SCITYPE")->asInt();
	} catch(...) {
		errstr<<"Unable to find manatory attribute `CCDMODE' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return st;
} // sciType()

unsigned short param(const vector<const Attributable *> &av)
{
	unsigned short p = 0;
	try {
		p = DalUtils::attribute(av, "PARAM")->asInt();
	} catch(...) {
		errstr<<"Unable to find manatory attribute `PARAM' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return p;
} // param()


//
//	access attribute F1294 ("FRAME_TIME_PARAMETER" of EPN) - possible
//	values [0, 3] or -1 (not present)
//
int frameTimeParameter(const vector<const Attributable *> &av)
{
	int ftp = -1;
	try {
		ftp = DalUtils::attribute(av, "F1294")->asInt();
	} catch (...) { }

	return ftp;
} // frameTimeParameter()

// SCR117
string filterUsed(const vector<const Attributable *> &av)
{
	string filter = "";
	try {
		filter = DalUtils::attribute(av, "FILTER")->asString();
	} catch (...) { }

	return filter;
} // filter()


}


//
//	---( Xmm )--------------------------------------------------------------
//
const STime Xmm::missionReferenceTime("1997-12-31T23:58:56.816"); // UTC
const STime Xmm::missionStartTime("1999-12-10T14:32:00.000");
const STime Xmm::missionEndTime(Xmm::missionStartTime + 10. *
								365. * STime::secondsPerDay);

const std::vector<Node> Xmm::emptyNodeVec;
const std::vector<Mode> Xmm::emptyModeVec;
const std::vector<DataMode> Xmm::emptyDataModeVec;
const std::vector<Filter> Xmm::emptyFilterVec;

static const char *__ins[] = { INSTRUMENT_STRINGS, 0 };
static const char *__insShort[] = { INSTRUMENT_SHORT_STRINGS, 0 };
static const StringPool _instruments(__ins, StringPool::ABORT);
static const StringPool _instrumentsShort(__insShort, StringPool::ABORT);

Xmm::Instrument Xmm::toInstrument(const string &insString)
{
	return S_CAST(toIntegralId((insString.length() > 2 ? _instruments
													   : _instrumentsShort),
							   insString, "instrument"), Xmm::Instrument);
} // Xmm::toInstrumentId()

const string &Xmm::toString(Instrument instr, bool shortForm)
{
	if (instr == Xmm::INSTRUMENT_NONE)
		return _undef;

	return !shortForm ? _instruments[UNT(instr)]
					  : _instrumentsShort[UNT(instr)];
} // Xmm::toString

ostream &operator<<(ostream &os, Instrument instrument)
{
	os<<INT(instrument)<<" ["<<Xmm::toString(instrument)<<"]";
	return os;
} // operator<<


//
//	nodes
//
static const char *__nodes[] = { NODE_STRINGS, 0 };
static const StringPool _nodes(__nodes, StringPool::ABORT);

Xmm::Node Xmm::toNode(const string &nodeString)
{
	return S_CAST(toIntegralId(_nodes, nodeString, "node"), Xmm::Node);
} // Xmm::toNode()

Xmm::Node Xmm::toNode(unsigned node)
{
	if (node > N_NODES)
		errstr<<"Cannot convert "<<node<<" to Node id - must be in range "
				"[0, "<<N_NODES<<"]\n";

	return S_CAST(node, Xmm::Node);
} // Xmm::toNode()

const string &Xmm::toString(Node node)
{
	if (node == Xmm::NODE_NONE)
		return _undef;

	return _nodes[UNT(node)];
} // Xmm::toString()

ostream &operator<<(ostream &os, Node node)
{
	os<<INT(node)<<" ["<<Xmm::toString(node)<<"]";
	return os;
} // operator<<


//
//	data modes
//
static const char *__modes[] = { DATA_MODE_STRINGS, 0 };
static const StringPool _dataModes(__modes, StringPool::ABORT);

Xmm::DataMode Xmm::toDataMode(const string &modeString)
{
	return S_CAST(toIntegralId(_dataModes, modeString, "data mode"),
				  Xmm::DataMode);
} // Xmm::toDataMode()


const string &Xmm::toString(DataMode dataMode)
{
	if (dataMode == Xmm::DATA_MODE_NONE)
		return _undef;

	return _dataModes[dataMode];
} // Xmm::toString()

ostream &operator<<(ostream &os, DataMode mode)
{
	os<<INT(mode)<<" ["<<Xmm::toString(mode)<<"]";
	return os;
} // operator<<


//
//	camera modes
//
static const char *__camModes[] = { MODE_STRINGS, 0 };
static const StringPool _camModes(__camModes, StringPool::ABORT);

Xmm::Mode Xmm::toMode(const string &modeString)
{
	return S_CAST(toIntegralId(_camModes, modeString, "camera mode"),
				  Xmm::Mode);
} // Xmm::toMode()

Xmm::Mode Xmm::toMode(unsigned mode)
{
	if (mode > N_MODES)
		errstr<<"Cannot convert "<<mode<<" to Mode id - must be in range "
				"[0, "<<N_MODES<<"]\n";

	return S_CAST(mode, Xmm::Mode);
} // Xmm::toMode()

const string &Xmm::toString(Mode mode)
{
	if (mode == Xmm::MODE_NONE)
		return _undef;

	return _camModes[mode];
} // Xmm::toString()


ostream &operator<<(ostream &os, Mode mode)
{
	os<<INT(mode)<<" ["<<Xmm::toString(mode)<<"]";
	return os;
} // operator<<


//
//	filters
//
static const char *__filters[] = { FILTER_STRINGS, 0 };
static const StringPool _filters(__filters, StringPool::ABORT);

Xmm::Filter Xmm::toFilter(const string &filterString)
{
	return S_CAST(toIntegralId(_filters, filterString, "filter"),
				  Xmm::Filter);
} // Xmm::toFilter()

Xmm::Filter Xmm::toFilter(unsigned filter)
{
	if (filter > N_FILTERS)
		errstr<<"Cannot convert "<<filter<<" to filter id - must be in range "
				"[0, "<<N_FILTERS<<"]\n";

	return S_CAST(filter, Xmm::Filter);
} // Xmm::toFilter()

const string &Xmm::toString(Filter filter)
{
	if (filter == Xmm::FILTER_NONE)
		return _undef;

	return _filters[filter];
} // Xmm::toString()

ostream &operator<<(ostream &os, Filter filter)
{
	os<<INT(filter)<<" ["<<Xmm::toString(filter)<<"]";
	return os;
} // operator<<


//
//	---( XmmInstrument )---------------------------------------------------
//
static DataMode __periodicHKmodes[] = { Xmm::HK_P1, Xmm::HK_P2, Xmm::HK_P3,
									    Xmm::HK_P4, Xmm::HK_P5, Xmm::HK_P6,
									    Xmm::HK_P7, Xmm::HK_P8, Xmm::HK_P9 };
static std::vector<DataMode> _periodicHKmodes(
					__periodicHKmodes,
					__periodicHKmodes + NUM_OF(__periodicHKmodes));

const std::vector<DataMode> &XmmInstrument::periodicHKmodes() const
{
	return _periodicHKmodes;
} // XmmInstrument::periodicHKmodes


static std::vector<DataMode> _scienceDataModes;	//	no science modes
const std::vector<DataMode> &XmmInstrument::scienceDataModes() const
{
	return _scienceDataModes;
} // XmmInstrument::scienceDataModes


static std::vector<DataMode> _scienceEngineeringDataModes; // no modes

const std::vector<DataMode> &XmmInstrument::scienceEngineeringDataModes() const
{
	return _scienceEngineeringDataModes;
} // XmmInstrument::scienceEngineeringDataModes


unsigned XmmInstrument::modeIndex(Mode mode) const
{
	const std::vector<Mode> &modes = this->modes();
	unsigned N = modes.size();
	for (unsigned n=0; n<N; ++n)
		if (mode == modes[n])
			return n;

	errstr<<"`"<<Xmm::toString(mode)<<"' is not a valid mode of "
		  <<Xmm::toString(id())<<Sas::error("UnknownMode");

	return 0;
} // XmmInstrument::modeIndex()


unsigned XmmInstrument::filterIndex(Filter filter) const
{
	const std::vector<Filter> &filters = this->filters();
	unsigned N = filters.size();
	for (unsigned n=0; n<N; ++n)
		if (filter == filters[n])
			return n;

	errstr<<"`"<<Xmm::toString(filter)<<"' is not a valid filter of "
		  <<Xmm::toString(id())<<Sas::error("UnknownFilter");

	return 0;
} // XmmInstrument::filterIndex()


unsigned short XmmInstrument::toHBR(Ccd) const
{
	errstr<<"Generic XmmInstrument cannot convert CCD to HBR number"
		  <<Sas::error("UnsupportedOperation");

	return 0;
} // XmmInstrument::toHBR()


//
//	-----( EMOS )----------------------------------------------------------
//
//	nodes
//
static Node __EMOSnodes[] = { Xmm::PRIMARY_NODE,
							  Xmm::REDUNDANT_NODE };
const std::vector<Node> EMOScam::nodesVec(__EMOSnodes,
									 __EMOSnodes+NUM_OF(__EMOSnodes));


//
//	camera modes
//
static const Mode __EMOScamModes[] = {
	Xmm::PRIME_FULL_WINDOW, Xmm::PRIME_PARTIAL_RFS, Xmm::PRIME_PARTIAL_W2,
	Xmm::PRIME_PARTIAL_W3, Xmm::PRIME_PARTIAL_W4, Xmm::PRIME_PARTIAL_W5,
	Xmm::PRIME_PARTIAL_W6, Xmm::FAST_UNCOMPRESSED, Xmm::FAST_COMPRESSED,
    Xmm::PRIME_FULL_WIN3X3,
	Xmm::OFFSET_VARIANCE, Xmm::CCD_DIAGNOSTIC,
	Xmm::DIAGNOSTIC_1X1, Xmm::DIAGNOSTIC_3X3,
	Xmm::EXTRA_HEATING_ANNEALING,
	Xmm::EXTRA_HEATING_DEICING, Xmm::EXTRA_HEATING_DECONTAM,
	Xmm::IN_FLIGHT_TEST, Xmm::PRIME_FULL_NO_BPMASK, Xmm::DIAGNOSTIC_1X1_RPP }; // SCR 117
const std::vector<Mode> EMOScam::modesVec(__EMOScamModes,
									 __EMOScamModes+NUM_OF(__EMOScamModes));

static const DataMode __EMOScamPeriodicHKmodes[] = {
	Xmm::PERIODIC_HOUSEKEEPING };
const std::vector<DataMode> EMOScam::periodicHKmodesVec(
				__EMOScamPeriodicHKmodes,
				__EMOScamPeriodicHKmodes + NUM_OF(__EMOScamPeriodicHKmodes));

static const DataMode __EMOScamScienceDataModes[] = {
	Xmm::IMAGING, Xmm::TIMING, Xmm::REDUCED_IMAGING, Xmm::COMPRESSED_TIMING
 };
const std::vector<DataMode> EMOScam::scienceDataModesVec(
				__EMOScamScienceDataModes,
				__EMOScamScienceDataModes + NUM_OF(__EMOScamScienceDataModes));

static const DataMode __EMOScamScienceEngineeringDataModes[] = {
	Xmm::IMAGING, Xmm::TIMING, Xmm::REDUCED_IMAGING, Xmm::COMPRESSED_TIMING,
	Xmm::DATA_DIAGNOSTIC, Xmm::DATA_OFFSET_VARIANCE
 };
const std::vector<DataMode> EMOScam::scienceEngineeringDataModesVec(
				__EMOScamScienceEngineeringDataModes,
				__EMOScamScienceEngineeringDataModes +
								NUM_OF(__EMOScamScienceEngineeringDataModes));

//
//	figure the camera mode from datatype, x0, y0, dx, dy
//
static bool isCloseAbs(unsigned val, unsigned nominalVal, unsigned tolerance)
{
	return UNT(abs(INT(val - nominalVal))) <= tolerance;
} // isCloseAbs()

namespace {

Mode defaultEmosMode(const vector<const Attributable *> &av, bool silent)
{
	Mode mode = xmmPayLoad.epn.defaultMode();
	vector<int> v = windowCoordinates(av);
	int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];

	if (!silent)
		errstr<<"Cannot uniquely determine EMOS camera mode from tuple: "<<endl
			  <<"   dataType   = "<<dataType(av)<<endl
			  <<"   x0/y0      = "<<x0<<"/"<<y0<<endl
			  <<"   dx/dy      = "<<dx<<"/"<<dy<<endl
			  <<"   frameTime  = "<<frameTime(av).first<<" [ms]"<<endl
			  <<"will continue with "<<mode<<" as effective "
				"mode - Please note: As a result of a possibly incorrect mode "
				"setting the retrieval of mode-dependent calibration "
				"quantities via the Cal is likely to yield INADEQUATE data "
				"which ultimately might lead to INCORRECT scientific results"
			  <<Sas::warning("UnidentifiedCameraMode");

	return mode;
} // defaultEmosMode()

}

pair<Mode, bool>
EMOScam::toMode(const vector<const Attributable *> &av,
				const std::vector<CoreModeParameters> &modeParameters,
				Ccd ccd, bool silent)
{
	pair<Mode, bool> mode(Xmm::MODE_NONE, true);
	unsigned dfModeOffset = UNT(xmmPayLoad.emos1.defaultMode());
	string dt = dataType(av);
	if (dt == "IMAGE.EL" || dt == "RIMAGE.EL") {
		//
		//	check for PRIME_FULL_NO_BPMASK
		//
		if (emdh(av).second == 32767) {
		  // 2003-06-19 GV: following the controversy on this new mode
		  // being introduced, I disable its detection here and issue
		  // a warning instead. This pending the resolution of the
		  // issue by the SAS CCB.
		  //
		  //`PrimeFullNoBadPixelMask' =
		  //- MOS/FF no BPmask
		  //* the exposure is in Full Frame Mode
		  //* EMDH=32767 (keyword in the header of the first extension
		  //in the above file; note that EMDH=4095 for "science"
		  //Full Frame Mode)

          // 2005-06-14 JSH: epframes does not have a 'pattern lib'
		  // for this mode, so it remains as is for the time being
		  // mode.first = Xmm::PRIME_FULL_NO_BPMASK;
		  // return mode;

		  Sas::warning("PossibleIncorrectSetting",
					   "I have detected an occurrence of the keyword EMDH=32767 instead of the expected value 4095. This may indicate that the instrument was operated with the wrong settings. In general it is ok to process this exposure as a normal FullFrame one.n");

		}

		//	get window coordinates
		vector<int> v = windowCoordinates(av);
		int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];

		//	frame time
		//	the following is ugly but unavoidable
		//
		pair<unsigned, string> ft = frameTime(av);
		if (ft.second.empty())
			//
			//	this is from the ODF in units of 100 ms
			//
			ft.first *= 100;
		else if (ft.second == "s" || ft.second == "sec")
			ft.first *= 1000;
		else if (ft.second != "ms")
			errstr<<"Unrecognized unit for attribute "
					"`FRMTIME\'"
				  <<Sas::error("UnrecognizedUnit");

		//
		//	image mode data
		//
		//	the mode is one of the following
		//
		//	ID MODE         X0,Y0,DX,DY        frameTime       EDUmode
		//	0  FULLWINDOW   0,0,600,600        2.6 ish         3
		//	1  PARTIAL RFS  0,0,600,600        0.2 ish         3
		//	2  W2           250,250,100,100    0.4 ish         3
		//	3  W3           150,150,300,300    0.9 ish         3
		//	4  W4           250,250,100,100    0.2 ish         3
		//	5  W5           150,150,300,300    0.7 ish         3
		//	6  W6           0,0,300,600        1.4 ish         3
		//	7  FULLWIN3x3   0,0,201,203     0.4, 0.6 or free-running  3
		//
		if (modeParameters.size() < 7) {
			errstr<<"Need at least mode parameters of EMOS camera modes:\n";
			for (unsigned n=0; n<7; ++n)
				errstr<<Xmm::toString(Xmm::toMode(n+dfModeOffset))<<endl;
			errstr<<"to determine effective instrument mode"
				  <<Sas::error("NoModeParameters");
		}
		std::vector<unsigned> modeIndices;
		for (unsigned n=0; n<modeParameters.size(); ++n)
			if (isCloseAbs(x0, modeParameters[n].windowX0, 5) &&
				isCloseAbs(y0, modeParameters[n].windowY0, 5) &&
				isCloseAbs(dx, modeParameters[n].windowDX, 10) &&
				isCloseAbs(dy, modeParameters[n].windowDY, 10) &&
				isCloseAbs(ft.first, modeParameters[n].frameTime, 100))
				//	found a match
				modeIndices.push_back(n);
		if (modeIndices.size() == 1)
			//	mode is unequivocally determined
			mode.first = Xmm::toMode(modeIndices[0] + dfModeOffset);
		else {
            // If no match found, check explicitely for 3x3 mode
            // where the frame time is a moveable feast

            unsigned n = Xmm::PRIME_FULL_WIN3X3;
            if ( modeIndices.size() == 0 &&
                 modeIndices.size() >= n &&
                isCloseAbs(x0, modeParameters[n].windowX0, 5) &&
                isCloseAbs(y0, modeParameters[n].windowY0, 5) &&
                isCloseAbs(dx, modeParameters[n].windowDX, 10) &&
                isCloseAbs(dy, modeParameters[n].windowDY, 10) )
                   mode.first = Xmm::PRIME_FULL_WIN3X3;
            else {
 			  //	could not figure the mode - if it is an outer CCD the mode
			  //	must be FullWindow
			  if (ccd > 1)
				mode.first = Xmm::PRIME_FULL_WINDOW;
			  else {
				if (modeIndices.size() > 1) {
					errstr<<"The camera mode is one of:\n";
					for (unsigned n=0; n<modeIndices.size(); ++n)
						errstr<<Xmm::toString(Xmm::toMode(modeIndices[n]
														+ dfModeOffset))<<endl;
				}

				mode.first = defaultEmosMode(av, silent);
				mode.second = false;
			  }
			}
		}
	} else if (dt == "TIME.EL")
		mode.first = Xmm::FAST_UNCOMPRESSED;
	else if (dt == "CTIME.EL")
		mode.first = Xmm::FAST_COMPRESSED;
	else if (dt == "OFFVAR.EL")
		mode.first = Xmm::OFFSET_VARIANCE;
	else if (dt == "DIAG.IM") {
		//	get window coordinates
		vector<int> v = windowCoordinates(av);
		int &dx = v[2], &dy = v[3];
		pair<unsigned, string> ft = frameTime(av);

		if (dx<600 && dy<600)
			mode.first = Xmm::DIAGNOSTIC_3X3;
		else if (dx>=600 && dy>=600)
		// SCR 117
		{
			if (ft.first == 80) {
				mode.first = Xmm::DIAGNOSTIC_1X1_RPP;
			}
			else {
				mode.first = Xmm::DIAGNOSTIC_1X1;
			}
		}
		else
			mode.first = Xmm::CCD_DIAGNOSTIC;
	}

	//
	//	the list of modes we have not checked for at this point is:
	//
	//		EXTRA_HEATING_ANNEALING
	//		EXTRA_HEATING_DEICING
	//		EXTRA_HEATING_DECONTAM
	//		IN_FLIGHT_TEST
	//
	else {
		mode.first = defaultEmosMode(av, silent);
		mode.second = false;
	}

	return mode;
} // EMOScam::toMode()


//
//	filters
//
static const Filter __EMOSfilters[] = {
	Xmm::OPEN, Xmm::CLOSED, Xmm::THIN1, Xmm::THIN2, Xmm::MEDIUM, Xmm::THICK,
	Xmm::CAL_OPEN, Xmm::CAL_CLOSED, Xmm::CAL_THIN1, Xmm::CAL_THIN2,
	Xmm::CAL_MEDIUM, Xmm::CAL_THICK };
const std::vector<Filter> EMOScam::filtersVec(__EMOSfilters,
									     __EMOSfilters+NUM_OF(__EMOSfilters));

unsigned short EMOScam::toHBR(Ccd ccd) const
{
	static bool initialized = false;
	static map<Ccd, unsigned short> ccdHBRmap;
	if (!initialized) {
		//
		//	initialzie CCD-HBR map according to following translation table
		//	(see email D. Lumb, dated 2001-03-21):
		//	HBR =   1       CCDNR = 1
		//	        3               2
		//	        7               3
		//	        5               4
		//	        4               5
		//	        8               6
		//	        6               7
		//
		ccdHBRmap[1] = 1;
		ccdHBRmap[2] = 3;
		ccdHBRmap[3] = 7;
		ccdHBRmap[4] = 5;
		ccdHBRmap[5] = 4;
		ccdHBRmap[6] = 8;
		ccdHBRmap[7] = 6;

		initialized = true;
	}

	map<Ccd, unsigned short>::const_iterator i = ccdHBRmap.find(ccd);
	if (i == ccdHBRmap.end())
		errstr<<"There is no HBR corresponding to CCD #"<<ccd
			  <<Sas::error("NoSuchHBR");

	return i->second;
} // EMOScam::toHBR()


//
//	-----( EPN )----------------------------------------------------------
//
//
//	camera modes
//
static const Mode __EPNcamModes[] = {
	Xmm::PRIME_FULL_WINDOW, Xmm::PRIME_FULL_WINDOW_EXTENDED,
	Xmm::PRIME_LARGE_WINDOW, Xmm::PRIME_SMALL_WINDOW,
	Xmm::FAST_TIMING, Xmm::FAST_BURST, Xmm::PRIME_FULL_OFFSET,
	Xmm::PRIME_LARGE_OFFSET, Xmm::OFFSET, Xmm::NOISE, Xmm::DIAGNOSTIC,
	Xmm::EXTRA_HEATING_DEICING, Xmm::EXTRA_HEATING_DECONTAM,
	Xmm::IN_FLIGHT_TEST };
const std::vector<Mode> EPNcam::modesVec(__EPNcamModes,
								    __EPNcamModes+NUM_OF(__EPNcamModes));

namespace {

Mode defaultEpnMode(const vector<const Attributable *> &av, bool silent)
{
	Mode mode = xmmPayLoad.epn.defaultMode();
	vector<int> v = windowCoordinates(av);
	int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];
	if (!silent)
		errstr<<"Cannot uniquely determine EPN camera mode from tuple: "
			  <<"   dataType   = "<<dataType(av)<<endl
			  <<"   x0/y0      = "<<x0<<"/"<<y0<<endl
			  <<"   dx/dy      = "<<dx<<"/"<<dy<<endl
			  <<"   frameTime  = "<<frameTime(av).first<<" [ms]"<<endl
			  <<"   ccdMode    = "<<ccdMode(av)<<endl
			  <<"will continue with "<<mode<<" as effective "
				"mode - Please note: As a result of a possibly incorrect mode "
				"setting the retrieval of mode-dependent calibration "
				"quantities via the Cal is likely to yield INAPPROPRIATE data"
				" which ultimately might lead to INCORRECT scientific results"
			  <<Sas::warning("UnidentifiedCameraMode");

	return mode;
} // defaultEpnMode()

}

pair<Mode, bool>
EPNcam::toMode(const vector<const Attributable *> &av,
			   const std::vector<CoreModeParameters> &modeParameters,
			   bool silent)
{
	pair<Mode, bool> mode(Xmm::MODE_NONE, true);
	string dt = dataType(av);
	if (dt == "IMAGE.EL") {
		pair<unsigned, string> ft = frameTime(av);
		unsigned ccdm = ccdMode(av);

		//
		//	image mode data
		//
		switch (ccdm) {
			case 1: //	this can be Xmm::PRIME_FULL_WINDOW_EXTENDED
					//	or Xmm::PRIME_FULL_WINDOW; decide through
					//	frame time
					if (modeParameters.size() < 2)
						errstr<<"Need at least mode parameters of EPN camera "
								"modes:\n"
							  <<Xmm::toString(Xmm::PRIME_FULL_WINDOW)
							  <<endl
							  <<Xmm::toString(Xmm::PRIME_FULL_WINDOW_EXTENDED)
							  <<"\nto determine effective instrument mode"
							  <<Sas::error("NoModeParameters");
					if (isCloseAbs(ft.first,
								   modeParameters[0].frameTime, 10))
						mode.first = Xmm::PRIME_FULL_WINDOW;
					else if (isCloseAbs(ft.first,
							 modeParameters[1].frameTime, 10))
						mode.first = Xmm::PRIME_FULL_WINDOW_EXTENDED;
					else {
						//	there was no close match in the frame time;
						//	this is probably PRIME_FULL_WINDOW_EXTENDED
						//	where PMS/ODS is known to fail to write the
						//	correct frame time into the FRAMETIME attribute;
						//	see if F1294 is present if yes and it is >0
						//	this is PRIME_FULL_WINDOW_EXTENDED
						int ftp = frameTimeParameter(av);
						if (ftp >= 0)
							mode.first = ftp > 0
											? Xmm::PRIME_FULL_WINDOW_EXTENDED
											: Xmm::PRIME_FULL_WINDOW;
						else {
							//	no F1294 attribute present - we give a warning
							//	and continue with
							//	PRIME_FULL_WINDOW_EXTENDED
							string effstr(
								Xmm::toString(
											Xmm::PRIME_FULL_WINDOW_EXTENDED));
							errstr<<"The window coordinates indicate this to "
									"be either `"
								  <<Xmm::toString(Xmm::PRIME_FULL_WINDOW)
								  <<"' or `"<<effstr
								  <<"' mode but the read frame time "
								  <<ft.first<<" [ms] does not match the "
									"nominal values ("
								  <<modeParameters[0].frameTime<<'/'
								  <<modeParameters[1].frameTime<<" [ms]) "
									"of either of those and the parameter "
									"F1293 (FRAME_TIME_PARAMETER) is not "
									"present - the frame time is likely "
									"to be the latter where the lack of an "
									"accurate FRAMETIME keyword for `"<<effstr
								  <<"' is a known ODF problem - will "
									"continue with `"<<effstr
								  <<"'. If you have created a summary file "
									"with odfingest you can check the value "
									"of FRAME_TIME_PARAMETER for this PN "
									"exposure. If it is not equal to 0 this "
									"is indeed an `"<<effstr<<"' exposure."
								  <<Sas::warning("EpnFrameTimeMismatch");
							mode.first = Xmm::PRIME_FULL_WINDOW_EXTENDED;
						}
					}
					break;
			case 2: mode.first = Xmm::PRIME_LARGE_WINDOW; break;
			case 3: mode.first = Xmm::PRIME_SMALL_WINDOW; break;
			default: break;
		}
	} else if (dt == "TIME.EL")
		mode.first = Xmm::FAST_TIMING;
	else if (dt == "BURST.EL")
		mode.first = Xmm::FAST_BURST;
	else if (dt == "DIAG.IM" || dt == "DISCLIN.IM")
		mode.first = Xmm::DIAGNOSTIC;
	else if (dt == "OFFSET.IM")
		mode.first = Xmm::OFFSET;
	else if (dt == "INTEGER*2")
		// SCR 117 - As of 11/07/05 - HDI Files have an incorrect DATATYPE field
		// In the future, they may be correct, in which case they will be handled above
		mode.first = Xmm::OFFSET;
	else if (dt == "NOISE.IM")
		mode.first = Xmm::NOISE;
	else {
		mode.first = defaultEpnMode(av, silent);
		mode.second = false;
	}

	return mode;
} // EPNcam::toMode()


static const DataMode __EPNcamPeriodicHKmodes[] = {
	Xmm::MAIN_PERIODIC_HOUSEKEEPING, Xmm::ADD_PERIODIC_HOUSEKEEPING };
const std::vector<DataMode> EPNcam::periodicHKmodesVec(
					__EPNcamPeriodicHKmodes,
					__EPNcamPeriodicHKmodes + NUM_OF(__EPNcamPeriodicHKmodes));

static const DataMode __EPNcamScienceDataModes[] = {
	Xmm::IMAGING, Xmm::TIMING, Xmm::BURST };
const std::vector<DataMode> EPNcam::scienceDataModesVec(
				__EPNcamScienceDataModes,
				__EPNcamScienceDataModes + NUM_OF(__EPNcamScienceDataModes));

static const DataMode __EPNcamScienceEngineeringDataModes[] = {
	Xmm::IMAGING, Xmm::TIMING, Xmm::BURST,
	Xmm::DATA_DIAGNOSTIC, Xmm::DATA_OFFSET_VARIANCE, Xmm::OFFSET_DATA,
	Xmm::NOISE_DATA, Xmm::DISCARDED_LINES_DATA
 };
const std::vector<DataMode> EPNcam::scienceEngineeringDataModesVec(
				__EPNcamScienceEngineeringDataModes,
				__EPNcamScienceEngineeringDataModes +
								NUM_OF(__EPNcamScienceEngineeringDataModes));

// RDS
static const DataMode __ERMcamScienceDataModes[] = {
        Xmm::ERM_COUNT_RATE, Xmm::ERM_SPECTRA };
const std::vector<DataMode> ERMcam::scienceDataModesVec(
                                __ERMcamScienceDataModes,
                                __ERMcamScienceDataModes + NUM_OF(__ERMcamScienceDataModes));
//
//	filters
//
static const Filter __EPNfilters[] = {
	Xmm::OPEN, Xmm::CLOSED, Xmm::THIN1, Xmm::THIN2,
	Xmm::MEDIUM, Xmm::THICK, Xmm::CAL_OPEN, Xmm::CAL_CLOSED,
	Xmm::CAL_THIN1, Xmm::CAL_THIN2, Xmm::CAL_MEDIUM,
	Xmm::CAL_THICK };
const std::vector<Filter> EPNcam::filtersVec(__EPNfilters,
									    __EPNfilters+NUM_OF(__EPNfilters));

unsigned EPNcam::toCcdId(unsigned ccd)
{
	if (ccd > 12)
		errstr<<"Illegal chip number - "<<ccd<<endl
			  <<Sas::error("IllegalChipNumber");

	return (ccd - 1)%3;
} // EPNcam::toCcdId()


unsigned EPNcam::toQuadrantId(unsigned ccd)
{
	if (ccd > 12)
		errstr<<"Illegal chip number - "<<ccd<<endl
			  <<Sas::error("IllegalChipNumber");

	return (ccd - 1)/3;
} // EPNcam::toQuadrantId()


unsigned EPNcam::toChipNr(unsigned ccdId, unsigned quadrantId)
{
	if (ccdId > 2)
		errstr<<"Illegal CCD number ("<<ccdId<<")"
			  <<Sas::error("IllegalCCDNumber");
	if (quadrantId > 3)
		errstr<<"Illegal quadrant number ("<<quadrantId<<")"
			  <<Sas::error("IllegalQuadrantNumber");

	return quadrantId * 3 + ccdId + 1;
} // EPNcam::toChipNr()

//
//	ccd		: [1, 12]
//	rawX	: [1, 64]
//	return	: quadrant column number [0, 191]
//
unsigned EPNcam::toQuadrantColumnNumber(Ccd ccd, unsigned rawX)
{
	return toQuadrantColumnNumber(toQuadrantId(ccd), toCcdId(ccd), rawX);
} // EPNcam::toQuadrantColumnNumber()


//
//	quadId: [0..3]
//	ccdId : [0..2]
//	rawX  : [1..64]
//	return: [0..191]
//
unsigned EPNcam::toQuadrantColumnNumber(unsigned quadId, unsigned ccdId,
									    unsigned rawX)
{
	if (!(quadId%2)) {	//	quadId == 0 || quadId == 2)
		return (2-ccdId)*64 + 64 - rawX;
	} else {
		return ccdId * 64 + 64 - rawX;
	}

	return 0;
} // EPNcam::toQuadrantColumnNumber()


//
//	-----( RGS )----------------------------------------------------------
//
//	nodes
//
static Node __RGSnodes[] = { Xmm::PRIMARY_NODE,
							 Xmm::REDUNDANT_NODE };
const std::vector<Node> RGScam::nodesVec(__RGSnodes,
								    __RGSnodes+NUM_OF(__RGSnodes));
static unsigned __RGSocbs[] = { 1, 2, 3, 4, 5 };
const std::vector<unsigned> RGScam::ocbs(__RGSocbs,
									__RGSocbs+NUM_OF(__RGSocbs));

//
//	camera modes
//
static const Mode __RGScamModes[] = {
	Xmm::SPECTROSCOPY, Xmm::HIGH_EVENT_RATE,
	Xmm::HIGH_EVENT_RATE_WITH_SES, Xmm::HIGH_EVENT_RATE_WITH_SER,
	Xmm::SPECTROSCOPY_SMALL_WINDOW,
	Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD,
	Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD,
	Xmm::DIAGNOSTIC_1X1, Xmm::DIAGNOSTIC_2X2, Xmm::DIAGNOSTIC_3X3,
	Xmm::DIAGNOSTIC_4X4, Xmm::DIAGNOSTIC_5X5, Xmm::SPECTROSCOPY_TEST,
	Xmm::DIAGNOSTIC_TEST };
const std::vector<Mode> RGScam::modesVec(__RGScamModes,
								    __RGScamModes+NUM_OF(__RGScamModes));

namespace {

Mode defaultRgsMode(const vector<const Attributable *> &av,
				    unsigned ocbFactor, bool silent)
{
	Mode mode = xmmPayLoad.rgs1.defaultMode();
	vector<int> v = windowCoordinates(av);
	int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];
	if (!silent)
		errstr<<"Cannot uniquely determine RGS camera mode from tuple: "
			  <<"   dataType   = "<<dataType(av)<<endl
			  <<"   x0/y0      = "<<x0<<"/"<<y0<<endl
			  <<"   dx/dy      = "<<dx<<"/"<<dy<<endl
			  <<"   ocbFactor  = "<<ocbFactor<<endl
			  <<"   sciType    = "<<sciType(av)<<endl
			  <<"will continue with "<<mode<<" as effective "
				"mode - Please note: As a result of a possibly incorrect mode "
				"setting the retrieval of mode-dependent calibration "
				"quantities via the Cal is likely to yield INAPPROPRIATE data "
				"which ultimately might lead to INCORRECT scientific results"
			  <<Sas::warning("UnidentifiedCameraMode");

	return mode;
} // defaultRgsMode()

}

pair<Mode, bool>
RGScam::toMode(const vector<const Attributable *> &av, unsigned ocbFactor,
			   bool silent)
{
	pair<Mode, bool> mode(Xmm::MODE_NONE, true);
	string dt = dataType(av);

	if (dt == "SPECT.EL") {
		//
		//	spectroscopy mode
		//
		unsigned scienceType = sciType(av);
		switch (scienceType) {
			case 1: mode.first = Xmm::SPECTROSCOPY; break;
			case 2: mode.first = Xmm::HIGH_EVENT_RATE; break;
			case 3: mode.first = Xmm::HIGH_EVENT_RATE_WITH_SER; break;
			case 4: mode.first = Xmm::HIGH_EVENT_RATE_WITH_SES; break;
			case 6: mode.first = Xmm::SPECTROSCOPY_SMALL_WINDOW; break;
			default:
				errstr<<"Illegal value of SCITYPE ("<<scienceType<<") in "
						"spectroscopy data mode"
					  <<Sas::error("IllegalAttributeValue"); break;
		}
	} else if (dt == "HTR.EL") {
		//
		//	this could be Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD or
		//	HIGH_TIME_RESOLUTION_MULTIPLE_CCD
		//	for the moment assume the second
		//
		unsigned scienceType = sciType(av);
		if (scienceType != 5)
			errstr<<"SCITYPE cannot be "<<scienceType
				  <<" in High Time Resolution data mode"
				  <<Sas::error("IllegalAttributeValue");
		mode.first = Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD;
	} else if (dt == "DIAG.IM")
		switch (ocbFactor) {
			case 1: mode.first = Xmm::DIAGNOSTIC_1X1; break;
			case 2: mode.first = Xmm::DIAGNOSTIC_2X2; break;
			case 3: mode.first = Xmm::DIAGNOSTIC_3X3; break;
			case 4: mode.first = Xmm::DIAGNOSTIC_4X4; break;
			case 5: mode.first = Xmm::DIAGNOSTIC_5X5; break;
			default: errstr<<"Illegal value of CCDOCB ("<<ocbFactor<<") for "
							 "diagnostic mode"
						   <<Sas::error("IllegalAttributeValue"); break;
		}
	else {
		mode.first = defaultRgsMode(av, ocbFactor, silent);
		mode.second = false;
	}

	return mode;
} // Rgscam::toMode()


static const DataMode __RGScamPeriodicHKmodes[] = {
	Xmm::FULL_PERIODIC_HOUSEKEEPING, Xmm::TEMP_PERIODIC_HOUSEKEEPING };
const std::vector<DataMode> RGScam::periodicHKmodesVec(
					__RGScamPeriodicHKmodes,
					__RGScamPeriodicHKmodes + NUM_OF(__RGScamPeriodicHKmodes));

static const DataMode __RGScamScienceDataModes[] = {
	Xmm::DATA_SPECTROSCOPY, Xmm::HIGH_TIME_RESOLUTION };
const std::vector<DataMode> RGScam::scienceDataModesVec(
				__RGScamScienceDataModes,
				__RGScamScienceDataModes + NUM_OF(__RGScamScienceDataModes));

static const DataMode __RGScamScienceEngineeringDataModes[] = {
	Xmm::DATA_SPECTROSCOPY, Xmm::HIGH_TIME_RESOLUTION,
	Xmm::DATA_DIAGNOSTIC
 };
const std::vector<DataMode> RGScam::scienceEngineeringDataModesVec(
				__RGScamScienceEngineeringDataModes,
				__RGScamScienceEngineeringDataModes +
								NUM_OF(__RGScamScienceEngineeringDataModes));

//
//	-----( OM )----------------------------------------------------------
//
//	nodes
//
static Node __OMnodes[] = { Xmm::PRIMARY_NODE,
							Xmm::REDUNDANT_NODE };
const std::vector<Node> OMcam::nodesVec(__OMnodes,
								   __OMnodes+NUM_OF(__OMnodes));

//
//	camera modes
//
static const Mode __OMcamModes[] = {
	Xmm::IMAGE, Xmm::FAST,
	Xmm::RAW_DATA_BLUE_DSP12, Xmm::RAW_DATA_BLUE_DSP1, Xmm::RAW_DATA_BLUE_DSP2,
	Xmm::CENTROIDING_DATA, Xmm::FULL_FRAME, Xmm::CENTROIDING_CONFIRMATION,
	Xmm::INTENSIFIER_CHARACTERISTICS, Xmm::FLATFIELDING,
	Xmm::DARK_LOW, Xmm::DARK_HIGH, Xmm::FLAT_FIELD_LOW, Xmm::FLAT_FIELD_HIGH };
const std::vector<Mode> OMcam::modesVec(__OMcamModes,
								   __OMcamModes+NUM_OF(__OMcamModes));

namespace {
Mode defaultOmMode(const vector<const Attributable *> &av, bool silent)
{
	Mode mode = xmmPayLoad.om.defaultMode();
	vector<int> v = windowCoordinates(av);
	int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];
	if (!silent)
		errstr<<"Cannot uniquely determine OM camera mode from tuple: "
			  <<"   dataType   = "<<dataType(av)<<endl
			  <<"   x0/y0      = "<<x0<<"/"<<y0<<endl
			  <<"   dx/dy      = "<<dx<<"/"<<dy<<endl
			  <<"will continue with "<<mode<<" as effective "
				"mode - Please note: As a result of a possibly incorrect mode "
				"setting the retrieval of mode-dependent calibration "
				"quantities via the Cal is likely to yield INAPPROPRIATE data "
				"which ultimately might lead to INCORRECT scientific results"
			  <<Sas::warning("UnidentifiedCameraMode");

	return mode;
} // defaultOmMode()

}

pair<Mode, bool>
OMcam::toMode(const vector<const Attributable *> &av, bool silent)
{
	pair<Mode, bool> mode(Xmm::MODE_NONE, true);
	string dt = dataType(av);
	string filter = filterUsed(av);

	if (dt == "IMG.IM")
		//
		//	image mode
		//
		// SCR 117
		if (filter == "1200") {
			// The CAL, not having access to HK data, cannot differentiate
			// between FLAT_FIELD_LOW and DARK_LOW (H5195 != 0)
			mode.first = Xmm::DARK_LOW;
		}
		else {
			mode.first = Xmm::IMAGE;
		}
	else if (dt == "FST.EL")
		//
		//	fast mode
		//
		mode.first = Xmm::FAST;
	else if (dt == "ENG.IM") {
		switch (param(av)) {
			// SCR 117
			case 3: mode.first = Xmm::CENTROIDING_DATA; break;
			case 4:
				if (filter == "1200") {
					// The CAL, not having access to HK data, cannot differentiate
					// between FLAT_FIELD_HIGH and DARK_HIGH (H5195 != 0)
					mode.first = Xmm::DARK_HIGH;
				}
				else {
					mode.first = Xmm::IMAGE;
				}
				break;
			case 5: mode.first = Xmm::CENTROIDING_CONFIRMATION; break;
			case 6: mode.first = Xmm::INTENSIFIER_CHARACTERISTICS; break;
			// SCR117, E3 and E7 are the same (E3 is the processed version of E7),
			// but E3 is corrupt onboard (see NCR-78)
			case 7: mode.first = Xmm::CENTROIDING_DATA; break;
			default: break;
		}
	} else {
		mode.first = defaultOmMode(av, silent);
		mode.second = false;
	}

	return mode;
} // OMcam::toMode()


static const DataMode __OMcamPeriodicHKmodes[] = {
	Xmm::PERIODIC_HOUSEKEEPING };
const std::vector<DataMode> OMcam::periodicHKmodesVec(
					__OMcamPeriodicHKmodes,
					__OMcamPeriodicHKmodes + NUM_OF(__OMcamPeriodicHKmodes));

static const DataMode __OMcamScienceDataModes[] = {
	Xmm::IMAGING, Xmm::DATA_FAST };
const std::vector<DataMode> OMcam::scienceDataModesVec(
					__OMcamScienceDataModes,
					__OMcamScienceDataModes + NUM_OF(__OMcamScienceDataModes));

static const DataMode __OMcamScienceEngineeringDataModes[] = {
	Xmm::IMAGING, Xmm::DATA_FAST,
	Xmm::DATA_DIAGNOSTIC, Xmm::ENGINEERING1, Xmm::ENGINEERING2,
	Xmm::ENGINEERING3, Xmm::ENGINEERING4, Xmm::ENGINEERING5, Xmm::ENGINEERING6,
	Xmm::ENGINEERING7, Xmm::ENGINEERING1
 };
const std::vector<DataMode> OMcam::scienceEngineeringDataModesVec(
				__OMcamScienceEngineeringDataModes,
				__OMcamScienceEngineeringDataModes +
								NUM_OF(__OMcamScienceEngineeringDataModes));

//
//	filters
//
static const Filter __OMfilters[] = {
	Xmm::BLOCKED, Xmm::UVW2, Xmm::UVM2, Xmm::UVW1, Xmm::U, Xmm::B, Xmm::V,
	Xmm::WHITE, Xmm::MAGNI, Xmm::GRISM1, Xmm::GRISM2,
    Xmm::GRISM10, Xmm::GRISM20, Xmm::BARRED_U };
const std::vector<Filter> OMcam::filtersVec(__OMfilters,
									   __OMfilters+NUM_OF(__OMfilters));


//
//	----( XmmPayLoad )-------------------------------------------------------
//
XmmInstrument XmmPayLoad::xmm;
EMOScam1 XmmPayLoad::emos1;
EMOScam2 XmmPayLoad::emos2;
ERMcam XmmPayLoad::erm;
EPNcam XmmPayLoad::epn;
RGScam1	XmmPayLoad::rgs1;
RGScam2 XmmPayLoad::rgs2;
OMcam XmmPayLoad::om;

const std::vector<XmmInstrument *> &XmmPayLoad::instruments()
{
	static std::vector<XmmInstrument *> _ins;
	static bool first = true;

	if (first) {
		first = false;

		_ins.push_back(&emos1);
		_ins.push_back(&emos2);
		_ins.push_back(&erm);
		_ins.push_back(&epn);
		_ins.push_back(&rgs1);
		_ins.push_back(&rgs2);
		_ins.push_back(&om);
	}

	return _ins;
} // XmmPayLoad::instruments()


const XmmInstrument &XmmPayLoad::instrument(Xmm::Instrument inst)
{
	switch(inst) {
		case Xmm::EMOS1	: return XmmPayLoad::emos1; break;
		case Xmm::EMOS2	: return XmmPayLoad::emos2; break;
		case Xmm::EPN	: return XmmPayLoad::epn; break;
		case Xmm::ERM	: return XmmPayLoad::erm; break;
		case Xmm::RGS1	: return XmmPayLoad::rgs1; break;
		case Xmm::RGS2	: return XmmPayLoad::rgs2; break;
		case Xmm::OM	: return XmmPayLoad::om; break;
		case Xmm::SC	: return XmmPayLoad::xmm; break;
		default:
			errstr<<"Unknown instrument `"<<toString(inst)<<"'"
				  <<Sas::error("UnknownInstrument");
			break;
	}

	return XmmPayLoad::xmm;
} // XmmPayLoad::instrument()


XmmOrbitSimulator::XmmOrbitSimulator(const STime &time)
		: KeplerOrbitSimulator()
{
	static const unsigned N_PAR = 6;

	double orbitalElem[][N_PAR+1] = {
		//
		//	NORAD TLE from http://oig1.gsfc.nasa.gov:
		//
		//	TIME					[yy.ddddd]
		//	inclination				[deg]
		//	R.A. of ascending node	[deg]
		//	eccentricity			[0<e<1]
		//	argument of perigee		[deg]
		//	mean anomaly			[deg]
		//	orbital period			[1/day]
		//
		{ 1999.+344.608/365.,
		  39.9041, 237.978, 0.8851996, 55.327, 357.9464, 0.55095466 },
		{ 1999.+344.6085/365.,
		  39.9041, 237.9777, 0.8851996, 55.327, 357.9464, 0.55095466 },
		{ 1999.+351.66667/365.,
		  38.931, 235.912, 0.794729, 56.927, 214.801, 0.50147251 },
		{ 1999.+356.43244/365.,
		  38.7261, 235.16, 0.7940363, 57.6308, 355.0195, 0.50144297 },
		{ 1999.+356.50779/365.,
		  39.003, 235.3762, 0.7951865, 57.5044, 8.8091, 0.50151554 },
		{ 1999.+358.42732/365.,
		  38.7641, 234.9416, 0.7938691, 57.9159, 355.185, 0.50143288 },
		{ 1999.+362.41508/365.,
		  38.6963, 234.6988, 0.7941813, 58.0903, 355.1654, 0.50143296 },
		{ 1999.+364.40884/365.,
		  38.7162, 234.4952, 0.7945439, 58.1824, 355.1173, 0.50142631 },
		{ 2000.+3.3970459/366.,
		  38.5478, 234.308, 0.795785, 58.0848, 355.0808, 0.50137751 },
		{ 2000.+9.5021638/366.,
		  38.488, 234.201, 0.7989415, 58.278, 17.177, 0.50136759 },
		{ 2000.+15.375/366.,
		  38.3625, 233.7652, 0.7983233, 58.5975, 357.3151, 0.50128056 },
		{ 2000.+23.493937/366.,
		  38.5466, 233.3665, 0.7984616, 59.1144, 23.0702, 0.50150066 },
		{ 2000.+25.375/366.,
		  38.0587, 232.8394, 0.7993743, 59.2953, 2.6531, 0.50132631 },
		{ 2000.+25.502164/366.,
		  38.071, 232.828, 0.7994795, 59.341, 25.734, 0.50130018 },
		{ 2000.+27.375/366.,
		  38.0311, 232.7572, 0.7994466, 59.3504, 3.7479, 0.50132179 },
		{ 2000.+31.502164/366.,
		  38.002, 232.518, 0.8002953, 59.58, 28.913, 0.50136388 },
		{ 2000.+35.333333/366.,
		  37.8677, 232.2021, 0.8006114, 59.8586, 0.0995, 0.50141219 },
		{ 2000.+37.333333/366.,
		  37.7364, 232.0443, 0.8003598, 60.0142, 1.0449, 0.5013854 },
		{ 2000.+41.333333/366.,
		  37.3779, 231.8996, 0.7993935, 60.1279, 2.9518, 0.50147606 },
		{ 2000.+43.333333/366.,
		  37.3698, 231.8231, 0.7993432, 60.1697, 4.0157, 0.5012105 },
		{ 2000.+47.291667/366.,
		  37.5675, 231.3228, 0.8002899, 60.5313, 358.2125, 0.501277 },
		{ 2000.+51.291667/366.,
		  37.436, 230.8269, 0.7997824, 60.9363, 0.0813, 0.50117407 },
		{ 2000.+51.464656/366.,
		  37.442, 230.806, 0.799864, 60.982, 32.292, 0.50131555 },
		{ 2000.+53.464656/366.,
		  37.429, 230.695, 0.7995385, 61.048, 33.404, 0.50128101 },
		{ 2000.+55.291667/366.,
		  37.4239, 230.621, 0.7995749, 61.0849, 3.2822, 0.50130176 },
		{ 2000.+57.291667/366.,
		  37.4115, 230.5403, 0.7998324, 61.1404, 4.3334, 0.50134666 },
		{ 2000.+61.370912/366.,
		  37.348, 230.288, 0.8003056, 61.351, 19.129, 0.5011945 },
		{ 2000.+63.25/366.,
		  37.3805, 230.0469, 0.800056, 61.552, 357.9089, 0.50121324 },
		{ 2000.+63.370912/366.,
		  37.304, 230.109, 0.800176, 61.512, 20, 0.50119891 },
		{ 2000.+67.370912/366.,
		  37.243, 229.803, 0.7992313, 61.76, 21.835, 0.50111691 },
		{ 2000.+69.25/366.,
		  37.2012, 229.7491, 0.7986935, 61.7478, 0.5393, 0.50102555 },
		{ 2000.+71.25/366.,
		  37.2156, 229.6837, 0.798731, 61.7588, 1.4662, 0.50101028 },
		{ 2000.+75.25/366.,
		  37.1209, 229.3574, 0.7994453, 61.995, 3.2458, 0.50117015 },
		{ 2000.+77.208333/366.,
		  37.106, 229.1269, 0.7991809, 62.194, 356.5584, 0.50111553 },
		{ 2000.+79.208333/366.,
		  37.0832, 228.9509, 0.7987042, 62.3325, 357.4163, 0.50101125 },
		{ 2000.+83.208333/366.,
		  37.1175, 228.8524, 0.7986812, 62.337, 359.195, 0.50099197 },
		{ 2000.+85.166667/366.,
		  37.1104, 228.7923, 0.799026, 62.3705, 352.5241, 0.5010438 },
		{ 2000.+87.166667/366.,
		  37.0882, 228.6897, 0.7993412, 62.454, 353.3297, 0.50108616 },
		{ 2000.+89.208333/366.,
		  37.0048, 228.5708, 0.7994546, 62.5329, 1.6093, 0.50109173 },
		{ 2000.+91.208333/366.,
		  36.9746, 228.4041, 0.7992925, 62.6817, 2.4309, 0.50109528 },
		{ 2000.+93.208333/366.,
		  36.9642, 228.2633, 0.7989348, 62.7869, 3.2582, 0.50103752 },
		{ 2000.+95.321464/366.,
		  36.943, 228.186, 0.7986343, 62.845, 24.539, 0.50095295 },
		{ 2000.+97.125/366.,
		  36.9867, 228.1404, 0.7983472, 62.8294, 350.0378, 0.50093285 },
		{ 2000.+99.125/366.,
		  36.9836, 228.1039, 0.7986819, 62.813, 350.9662, 0.50100094 },
		{ 2000.+99.166667/366.,
		  36.9559, 228.1129, 0.7986775, 62.7962, 358.4774, 0.5009989 },
		{ 2000.+101.125/366.,
		  36.9162, 227.9875, 0.7992767, 62.8721, 351.839, 0.50109291 },
		{ 2000.+103.16667/366.,
		  36.8287, 227.7806, 0.7995101, 63.0333, 0.2043, 0.50110911 },
		{ 2000.+105.125/366.,
		  36.7886, 227.5942, 0.7992909, 63.1721, 353.5688, 0.50104045 },
		{ 2000.+107.16667/366.,
		  36.801, 227.4861, 0.7990461, 63.2259, 2, 0.50097198 },
		{ 2000.+109.32146/366.,
		  36.81, 227.466, 0.7992524, 63.228, 30.854, 0.50094705 },
		{ 2000.+111.125/366.,
		  36.8169, 227.4384, 0.7995471, 63.1872, 356.2957, 0.501006 },
		{ 2000.+113.125/366.,
		  36.8747, 227.3221, 0.8000756, 63.2708, 357.1968, 0.50106155 },
		{ 2000.+117.125/366.,
		  36.8421, 227.0794, 0.8007979, 63.4395, 358.9336, 0.50113684 },
		{ 2000.+117.32146/366.,
		  36.743, 227.131, 0.8008221, 63.474, 34.165, 0.50106287 },
		{ 2000.+119.125/366.,
		  36.6435, 227.0575, 0.8005921, 63.5035, 359.836, 0.50113735 },
		{ 2000.+123.125/366.,
		  36.6563, 226.8137, 0.8001154, 63.6762, 1.7508, 0.50104238 },
		{ 2000.+125.08333/366.,
		  36.6584, 226.7875, 0.8002517, 63.6526, 355.2268, 0.50102058 },
		{ 2000.+127.125/366.,
		  36.6456, 226.7194, 0.8008932, 63.6806, 3.7402, 0.50111396 },
		{ 2000.+127.32146/366.,
		  36.664, 226.702, 0.8011391, 63.712, 38.518, 0.50093654 },
		{ 2000.+129.08333/366.,
		  36.5881, 226.5428, 0.801611, 63.8104, 357.144, 0.50120803 },
		{ 2000.+131.08333/366.,
		  36.4558, 226.3234, 0.8017728, 63.9794, 358.0361, 0.5011835 },
		{ 2000.+133.04167/366.,
		  36.4243, 226.1599, 0.8016405, 64.0976, 351.4675, 0.50110655 },
		{ 2000.+135.08333/366.,
		  36.5088, 226.0798, 0.8017049, 64.1203, 0.1902, 0.50111023 },
		{ 2000.+137.08333/366.,
		  36.583, 226.0148, 0.8021499, 64.1095, 1.2081, 0.50114891 },
		{ 2000.+137.16718/366.,
		  36.495, 226.039, 0.8022281, 64.15, 16.127, 0.50102737 },
		{ 2000.+140.29167/366.,
		  36.5596, 225.9128, 0.8030731, 64.1763, 220.2127, 0.50125836 },
		{ 2000.+141.04167/366.,
		  36.549, 225.868, 0.803397, 64.2141, 355.5543, 0.5012369 },
		{ 2000.+143.04167/366.,
		  36.5112, 225.7274, 0.803892, 64.3419, 356.4359, 0.50125792 },
		{ 2000.+145.04167/366.,
		  36.4176, 225.5841, 0.8040191, 64.4977, 357.4482, 0.50129116 },
		{ 2000.+147/366.,
		  36.3802, 225.4222, 0.804012, 64.6412, 350.8493, 0.50125749 },
		{ 2000.+149/366.,
		  36.3557, 225.2908, 0.8038582, 64.7452, 351.8096, 0.50120228 },
		{ 2000.+153.16718/366.,
		  36.301, 225.156, 0.8043142, 64.838, 23.316, 0.5010127 },
		{ 2000.+155.04167/366.,
		  36.2616, 225.0582, 0.804945, 64.835, 2.5223, 0.50125411 },
		{ 2000.+157/366.,
		  36.1878, 224.8071, 0.8055942, 65.0432, 355.9755, 0.50135303 },
		{ 2000.+159/366.,
		  36.1382, 224.5319, 0.8056135, 65.3019, 357.0888, 0.50132379 },
		{ 2000.+160.95833/366.,
		  36.1106, 224.3704, 0.8055044, 65.4214, 350.6207, 0.50124966 },
		{ 2000.+162.95833/366.,
		  36.1039, 224.29, 0.8056549, 65.4657, 351.6763, 0.50123316 },
		{ 2000.+163/366.,
		  36.156, 224.3098, 0.8056539, 65.461, 359.276, 0.50125662 },
		{ 2000.+164.95833/366.,
		  36.1462, 224.2444, 0.8061194, 65.499, 352.7979, 0.50129402 },
		{ 2000.+166.95833/366.,
		  36.1231, 224.1439, 0.8067451, 65.5829, 353.7923, 0.50134726 },
		{ 2000.+167/366.,
		  36.0897, 224.1157, 0.8067529, 65.5609, 1.3666, 0.50134555 },
		{ 2000.+168.95833/366.,
		  36.0514, 223.9667, 0.8073313, 65.6991, 354.8164, 0.50138897 },
		{ 2000.+170.83333/366.,
		  36.0038, 223.7822, 0.8077009, 65.8747, 333.2025, 0.50142131 },
		{ 2000.+172.95833/366.,
		  35.9578, 223.5719, 0.8078121, 66.0758, 356.7205, 0.50139353 },
		{ 2000.+174.95833/366.,
		  35.879, 223.3833, 0.8076102, 66.2991, 357.949, 0.5014066 },
		{ 2000.+176.91667/366.,
		  35.8531, 223.2371, 0.807401, 66.4201, 351.4893, 0.50135031 },
		{ 2000.+178.91667/366.,
		  35.8385, 223.1386, 0.8073622, 66.4834, 352.5957, 0.50130776 },
		{ 2000.+180.95833/366.,
		  35.8385, 223.1023, 0.8073622, 66.5347, 1.0569, 0.50130776 },
		{ 2000.+182.91667/366.,
		  35.7634, 222.8644, 0.8084816, 66.646, 355.0149, 0.50150587 },
		{ 2000.+184.91667/366.,
		  35.7634, 222.8284, 0.8084816, 66.6969, 356.0991, 0.50150587 },
		{ 2000.+186.91667/366.,
		  35.6266, 222.2629, 0.8087164, 67.1651, 357.2079, 0.50145793 },
		{ 2000.+190.91667/366.,
		  35.5786, 222.0098, 0.8084962, 67.3948, 359.6986, 0.50141836 },
		{ 2000.+192.875/366.,
		  35.5616, 221.9094, 0.8089106, 67.4737, 353.2986, 0.5014666 },
		{ 2000.+194.91667/366.,
		  35.6211, 221.7588, 0.8095907, 67.5589, 1.9925, 0.50152834 },
		{ 2000.+196.875/366.,
		  35.579, 221.5669, 0.8099578, 67.7428, 355.5319, 0.50156255 },
		{ 2000.+199.87529/366.,
		  35.39, 221.295, 0.8202748, 67.394, 68.236, 0.45803221 },
		{ 2000.+202.875/366.,
		  35.465, 220.9768, 0.8096681, 68.3105, 358.7599, 0.50155724 },
		{ 2000.+204.83333/366.,
		  35.3717, 220.8218, 0.8092837, 68.3777, 352.8717, 0.50151389 },
		{ 2000.+206.83333/366.,
		  35.3597, 220.7207, 0.8091956, 68.4439, 354.1095, 0.50149576 },
		{ 2000.+208.83333/366.,
		  35.3348, 220.6, 0.8095325, 68.5257, 355.3497, 0.50154203 },
		{ 2000.+209.125/366.,
		  35.3384, 220.5751, 0.8094202, 68.6072, 48.3826, 0.50157049 },
		{ 2000.+210.83333/366.,
		  35.2899, 220.3448, 0.8099094, 68.7997, 356.9688, 0.50171918 },
		{ 2000.+212.83333/366.,
		  35.22, 220.0078, 0.8099653, 69.0959, 358.1984, 0.50172086 },
		{ 2000.+214.83333/366.,
		  35.1811, 219.7413, 0.809543, 69.3192, 359.4739, 0.50162491 },
		{ 2000.+216.79167/366.,
		  35.1727, 219.6054, 0.8092469, 69.4144, 353.2587, 0.50157382 },
		{ 2000.+218.16667/366.,
		  35.1513, 219.5174, 0.8092448, 69.4373, 241.7574, 0.5017871 },
		{ 2000.+218.79167/366.,
		  35.1527, 219.4924, 0.8092717, 69.4539, 354.6628, 0.50161767 },
		{ 2000.+220.79167/366.,
		  35.1371, 219.3797, 0.8095826, 69.5464, 355.9158, 0.50165962 },
		{ 2000.+222.79167/366.,
		  35.1828, 219.2545, 0.8099713, 69.708, 357.2469, 0.50173311 },
		{ 2000.+224.75/366.,
		  35.1469, 219.053, 0.8101069, 69.9004, 350.9239, 0.50174706 },
		{ 2000.+226.75/366.,
		  35.1104, 218.8327, 0.8099836, 70.1114, 352.1246, 0.50172316 },
		{ 2000.+228.75/366.,
		  35.0824, 218.624, 0.809594, 70.3052, 353.346, 0.501706 },
		{ 2000.+230.75/366.,
		  35.0261, 218.4344, 0.8090714, 70.4584, 354.707, 0.50167039 },
		{ 2000.+232.75/366.,
		  35.0199, 218.327, 0.8086504, 70.5341, 356.0299, 0.50162703 },
		{ 2000.+234.75/366.,
		  35.0139, 218.2474, 0.8085702, 70.5759, 357.38, 0.5016198 },
		{ 2000.+236.75/366.,
		  34.9945, 218.1226, 0.8088777, 70.6615, 358.7352, 0.50168805 },
		{ 2000.+238.70833/366.,
		  34.9455, 217.8719, 0.8091571, 70.872, 352.5081, 0.50177867 },
		{ 2000.+240.66667/366.,
		  34.8969, 217.5659, 0.8089376, 71.1347, 346.28, 0.50173016 },
		{ 2000.+242.75/366.,
		  34.8641, 217.388, 0.8084395, 71.2666, 2.6934, 0.50164543 },
		{ 2000.+244.70833/366.,
		  34.8671, 217.3009, 0.8081433, 71.3112, 356.5428, 0.50160977 },
		{ 2000.+248.70833/366.,
		  34.8575, 217.1389, 0.8085144, 71.4186, 359.1752, 0.50170281 },
		{ 2000.+248.74755/366.,
		  34.848, 217.1, 0.8087315, 71.469, 6.126, 0.50140772 },
		{ 2000.+250.70833/366.,
		  34.8355, 216.9899, 0.8087326, 71.5527, 0.4268, 0.50173239 },
		{ 2000.+252.70833/366.,
		  34.7781, 216.8053, 0.8086929, 71.7222, 1.4741, 0.5017059 },
		{ 2000.+254.66667/366.,
		  34.7537, 216.6164, 0.8084381, 71.8968, 355.1544, 0.50168686 },
		{ 2000.+256.66667/366.,
		  34.8099, 216.4163, 0.8080838, 72.0784, 356.2547, 0.50159142 },
		{ 2000.+260.625/366.,
		  34.8081, 216.2492, 0.8072219, 72.1783, 351.2802, 0.50150868 },
		{ 2000.+262.625/366.,
		  34.8044, 216.1959, 0.8072921, 72.1878, 352.5902, 0.50154244 },
		{ 2000.+266.66667/366.,
		  34.6513, 215.8326, 0.8076565, 72.4679, 2.4262, 0.50162117 },
		{ 2000.+270.625/366.,
		  34.6169, 215.4569, 0.8068803, 72.7522, 357.4133, 0.50145703 },
		{ 2000.+272.625/366.,
		  34.6492, 215.4829, 0.8066905, 72.768, 358.3334, 0.5013861 },
		{ 2000.+274.58333/366.,
		  34.6515, 215.4457, 0.8069463, 72.7714, 352.0062, 0.50143761 },
		{ 2000.+276.58333/366.,
		  34.6407, 215.359, 0.8073241, 72.8346, 353.1484, 0.50149475 },
		{ 2000.+278.58333/366.,
		  34.621, 215.2225, 0.8075973, 72.954, 354.2556, 0.50151407 },
		{ 2000.+280.625/366.,
		  34.5933, 215.0923, 0.8078816, 72.9898, 2.6558, 0.5014752 },
		{ 2000.+282.58333/366.,
		  34.5748, 214.933, 0.8076674, 73.1307, 356.2085, 0.50145187 },
		{ 2000.+284.58333/366.,
		  34.5656, 214.8071, 0.8072842, 73.2311, 357.3162, 0.50139874 },
		{ 2000.+288.58333/366.,
		  34.6472, 214.5744, 0.8067898, 73.355, 359.5417, 0.50129388 },
		{ 2000.+290.54167/366.,
		  34.6373, 214.52, 0.8071448, 73.3636, 353.1988, 0.50136838 },
		{ 2000.+292.58333/366.,
		  34.5496, 214.4625, 0.8077382, 73.3959, 1.7493, 0.50142961 },
		{ 2000.+294.54167/366.,
		  34.5096, 214.2147, 0.8078791, 73.5997, 355.329, 0.50143857 },
		{ 2000.+296.54167/366.,
		  34.4905, 214.0166, 0.8076176, 73.7516, 356.4618, 0.50136115 },
		{ 2000.+300.54167/366.,
		  34.5103, 213.8481, 0.8076713, 73.8101, 358.6054, 0.50127501 },
		{ 2000.+302.5/366.,
		  34.5053, 213.8018, 0.8081617, 73.8259, 352.1799, 0.50134033 },
		{ 2000.+304.5/366.,
		  34.4878, 213.6971, 0.8087108, 73.9095, 353.2246, 0.50137802 },
		{ 2000.+308.54167/366.,
		  34.4824, 213.3497, 0.8095186, 74.1258, 2.6309, 0.50137581 },
		{ 2000.+310.5/366.,
		  34.4626, 213.1905, 0.8094165, 74.2658, 356.1203, 0.50133984 },
		{ 2000.+312.5/366.,
		  34.452, 213.0705, 0.8091897, 74.3595, 357.166, 0.50129275 },
		{ 2000.+314.45833/366.,
		  34.3994, 213.0398, 0.8087724, 74.4714, 350.6084, 0.50121284 },
		{ 2000.+316.45833/366.,
		  34.3938, 212.9993, 0.8089893, 74.4687, 351.7067, 0.50120991 },
		{ 2000.+318.45833/366.,
		  34.3708, 212.8947, 0.8096084, 74.5287, 352.7899, 0.50130308 },
		{ 2000.+320.5/366.,
		  34.3033, 212.6322, 0.8101934, 74.7303, 1.2549, 0.50134918 },
		{ 2000.+322.45833/366.,
		  34.2636, 212.3557, 0.8102921, 74.9635, 354.7528, 0.5013197 },
		{ 2000.+322.64243/366.,
		  34.3, 212.364, 0.8104332, 74.99, 29.099, 0.50136131 },
		{ 2000.+324.45833/366.,
		  34.2487, 212.1743, 0.8101303, 75.1014, 355.8097, 0.50123609 },
		{ 2000.+328.45833/366.,
		  34.2639, 212.0338, 0.8106279, 75.1645, 357.8788, 0.50122631 },
		{ 2000.+330.41667/366.,
		  34.2477, 211.9361, 0.8112665, 75.2389, 351.3551, 0.50127425 },
		{ 2000.+332.41667/366.,
		  34.2221, 211.7757, 0.8118603, 75.3846, 352.3047, 0.50133319 },
		{ 2000.+334.45833/366.,
		  34.3295, 211.4427, 0.8125538, 75.7571, 0.7135, 0.50132984 },
		{ 2000.+336.41667/366.,
		  34.301, 211.2315, 0.8126706, 75.955, 354.1182, 0.50131558 },
		{ 2000.+338.41667/366.,
		  34.2797, 211.0422, 0.812574, 76.1273, 355.0655, 0.50127782 },
		{ 2000.+340.41667/366.,
		  34.2648, 210.8912, 0.8123993, 76.2499, 356.05, 0.50127102 },
		{ 2000.+342.41667/366.,
		  34.1011, 210.951, 0.8120196, 76.097, 356.9803, 0.50118687 },
		{ 2000.+344.375/366.,
		  34.0861, 210.8617, 0.8123692, 76.1468, 350.5027, 0.5012132 },
		{ 2000.+346.375/366.,
		  34.0511, 210.6755, 0.8130479, 76.2913, 351.5186, 0.50130809 },
		{ 2000.+348.41667/366.,
		  34.0016, 210.3506, 0.8135125, 76.5749, 0.0136, 0.50135437 },
		{ 2000.+350.375/366.,
		  33.9673, 210.0366, 0.8134366, 76.8469, 353.4832, 0.50128053 },
		{ 2000.+352.375/366.,
		  33.9562, 209.8518, 0.8132603, 76.9915, 354.5129, 0.50120463 },
		{ 2000.+354.375/366.,
		  33.9513, 209.749, 0.8134047, 77.0599, 355.5425, 0.50120114 },
		{ 2000.+356.375/366.,
		  33.939, 209.6388, 0.813868, 77.1451, 356.5338, 0.50124514 },
		{ 2000.+358.33333/366.,
		  33.9654, 209.4126, 0.8150084, 77.4668, 348.163, 0.50114275 },
		{ 2000.+360.375/366.,
		  33.9368, 209.1885, 0.8154493, 77.681, 356.4726, 0.50117474 },
		{ 2000.+362.33333/366.,
		  33.9076, 208.9349, 0.815637, 77.9258, 349.7327, 0.50117563 },
		{ 2000.+364.25985/366.,
		  33.883, 208.6872, 0.8155563, 78.1627, 337.2626, 0.50114789 },
		{ 2001.+2.375/365.,
		  33.8384, 208.3574, 0.8146676, 78.3695, 358.9938, 0.50089489 },
		{ 2001.+4.3333333/365.,
		  33.8289, 208.24, 0.8145958, 78.4574, 352.2535, 0.5008785 },
		{ 2001.+6.3333333/365.,
		  33.8109, 208.1021, 0.8148874, 78.5593, 353.0385, 0.50092954 },
		{ 2001.+8.3333333/365.,
		  33.7742, 207.8462, 0.8153944, 78.7758, 353.7949, 0.50101071 },
		{ 2001.+14.333333/365.,
		  33.8801, 206.896, 0.8154242, 79.7423, 355.91, 0.50083679 },
		{ 2001.+16.333333/365.,
		  33.8773, 206.7835, 0.8155123, 79.823, 356.6708, 0.5008645 },
		{ 2001.+18.291667/365.,
		  33.8657, 206.6403, 0.8158531, 79.9441, 349.8699, 0.50092126 },
		{ 2001.+20.291667/365.,
		  33.8468, 206.4377, 0.8161959, 80.1329, 350.5361, 0.50096318 },
		{ 2001.+22.333333/365.,
		  33.7286, 206.2663, 0.8162013, 80.2859, 358.7267, 0.50097988 },
		{ 2001.+24.291667/365.,
		  33.7099, 206.0055, 0.8160851, 80.5379, 351.8483, 0.50097076 },
		{ 2001.+28.333333/365.,
		  33.7099, 205.9254, 0.8160851, 80.6563, 0.7607, 0.50097076 },
		{ 2001.+30.291667/365.,
		  33.6205, 205.4221, 0.8141308, 80.838, 354.1286, 0.50087965 },
		{ 2001.+32.291667/365.,
		  33.6161, 205.3166, 0.8139808, 80.9122, 354.9178, 0.50088506 },
		{ 2001.+34.333333/365.,
		  33.6044, 205.1706, 0.814324, 81.0735, 3.1939, 0.50093318 },
		{ 2001.+36.291667/365.,
		  33.5761, 204.8947, 0.8145725, 81.3104, 356.4356, 0.50101712 },
		{ 2001.+40.291667/365.,
		  33.5294, 204.2902, 0.8138327, 81.8454, 357.9643, 0.50088651 },
		{ 2001.+42.25/365.,
		  33.5359, 204.1758, 0.8134592, 81.9174, 351.2701, 0.50084295 },
		{ 2001.+44.25/365.,
		  33.5368, 204.0881, 0.8135166, 81.9725, 352.0578, 0.50088342 },
		{ 2001.+46.291667/365.,
		  33.5396, 203.9467, 0.8137688, 82.1117, 0.4586, 0.5009775 },
		{ 2001.+48.25/365.,
		  33.527, 203.7546, 0.813933, 82.2892, 353.6591, 0.50101217 },
		{ 2001.+50.25/365.,
		  33.5152, 203.5309, 0.8138714, 82.5022, 354.3547, 0.50100834 },
		{ 2001.+52.25/365.,
		  33.4962, 203.2922, 0.8134944, 82.7137, 355.1238, 0.50101085 },
		{ 2001.+54.25/365.,
		  33.4939, 203.1098, 0.8129708, 82.8805, 355.8559, 0.5009531 },
		{ 2001.+56.208333/365.,
		  33.4955, 202.9859, 0.812401, 82.9827, 349.1128, 0.50089239 },
		{ 2001.+58.208333/365.,
		  33.4987, 202.9113, 0.8119861, 83.0276, 349.9214, 0.50087507 },
		{ 2001.+60.25/365.,
		  33.5386, 202.8369, 0.8119258, 83.0616, 358.536, 0.50094939 },
		{ 2001.+62.208333/365.,
		  33.5252, 202.6948, 0.8121089, 83.165, 351.874, 0.50102625 },
		{ 2001.+64.208333/365.,
		  33.5066, 202.4335, 0.8122073, 83.386, 352.6963, 0.50107469 },
		{ 2001.+66.25/365.,
		  33.4618, 202.1407, 0.8117686, 83.6642, 1.3008, 0.50108013 },
		{ 2001.+68.208333/365.,
		  33.4676, 201.9733, 0.8111614, 83.7843, 354.6988, 0.50096642 },
		{ 2001.+70.208333/365.,
		  33.4767, 201.9207, 0.8108981, 83.7919, 355.6342, 0.5009541 },
		{ 2001.+72.208333/365.,
		  33.4774, 201.8659, 0.811067, 83.8116, 356.5296, 0.50100986 },
		{ 2001.+74.208333/365.,
		  33.4627, 201.7503, 0.8113635, 83.9048, 357.371, 0.50107273 },
		{ 2001.+76.166667/365.,
		  33.4536, 201.5817, 0.8115154, 84.0564, 350.6552, 0.5010972 },
		{ 2001.+78.166667/365.,
		  33.4743, 201.3741, 0.8117538, 84.3036, 351.6585, 0.50109181 },
		{ 2001.+80.208333/365.,
		  33.4699, 201.1931, 0.8114152, 84.4642, 359.9729, 0.50111802 },
		{ 2001.+82.166667/365.,
		  33.472, 201.0691, 0.8109475, 84.5671, 353.2862, 0.50101236 },
		{ 2001.+84.166667/365.,
		  33.4751, 201.0012, 0.8104938, 84.608, 354.1552, 0.50096369 },
		{ 2001.+86.166667/365.,
		  33.4761, 200.9701, 0.8102815, 84.6036, 355.0579, 0.50095776 },
		{ 2001.+88.166667/365.,
		  33.4572, 200.9385, 0.8101015, 84.5113, 356.2022, 0.50103989 },
		{ 2001.+90.166667/365.,
		  33.4414, 200.7835, 0.810419, 84.6253, 357.1186, 0.50111399 },
		{ 2001.+92.166667/365.,
		  33.4277, 200.5315, 0.8104859, 84.8359, 358.0091, 0.50114759 },
		{ 2001.+94.166667/365.,
		  33.425, 200.3019, 0.8100495, 85.0393, 358.4411, 0.50100069 },
		{ 2001.+96.125/365.,
		  33.4329, 200.2036, 0.8096501, 85.089, 351.8372, 0.50092065 },
		{ 2001.+98.125/365.,
		  33.4373, 200.181, 0.8096722, 85.068, 352.7531, 0.50092686 },
		{ 2001.+100.125/365.,
		  33.4329, 200.1279, 0.8100607, 85.0873, 353.6213, 0.50100902 },
		{ 2001.+102.125/365.,
		  33.4213, 200.0037, 0.810516, 85.1909, 354.4381, 0.50105089 },
		{ 2001.+104.125/365.,
		  33.4108, 199.8325, 0.8107634, 85.3451, 355.2182, 0.50106719 },
		{ 2001.+106.125/365.,
		  33.4035, 199.6536, 0.81074, 85.5087, 355.993, 0.50105884 },
		{ 2001.+108.125/365.,
		  33.401, 199.4955, 0.8104474, 85.6558, 357.367, 0.50114066 },
		{ 2001.+110.08333/365.,
		  33.4004, 199.3942, 0.8101188, 85.7342, 350.764, 0.50109051 },
		{ 2001.+112.08333/365.,
		  33.4003, 199.3438, 0.8098909, 85.7537, 351.7167, 0.50106044 },
		{ 2001.+114.08333/365.,
		  33.3642, 199.2718, 0.8100216, 85.7431, 351.1273, 0.50092644 },
		{ 2001.+116.125/365.,
		  33.3499, 199.1795, 0.8104365, 85.7946, 359.5144, 0.50098716 },
		{ 2001.+118.08333/365.,
		  33.3313, 198.9708, 0.8108757, 85.9621, 352.8387, 0.50105519 },
		{ 2001.+120.08333/365.,
		  33.3211, 198.7011, 0.8109313, 86.191, 353.6633, 0.50104019 },
		{ 2001.+122.125/365.,
		  33.3224, 198.5074, 0.8106588, 86.3412, 2.0331, 0.50095311 },
		{ 2001.+124.08333/365.,
		  33.326, 198.4327, 0.8105344, 86.3719, 355.4075, 0.50090521 },
		{ 2001.+126.08333/365.,
		  33.3237, 198.3907, 0.8108411, 86.3753, 356.2832, 0.50093589 },
		{ 2001.+128.08333/365.,
		  33.3139, 198.295, 0.8114223, 86.4425, 357.1071, 0.50099655 },
		{ 2001.+130.08333/365.,
		  33.3019, 198.1273, 0.8119538, 86.5915, 357.8793, 0.50105798 },
		{ 2001.+132.04167/365.,
		  33.2903, 197.9213, 0.812245, 86.7829, 351.1094, 0.50106548 },
		{ 2001.+134.04167/365.,
		  33.2825, 197.7181, 0.8122602, 86.9727, 351.8576, 0.5010416 },
		{ 2001.+136.08333/365.,
		  33.2576, 197.6198, 0.8103524, 86.6227, 2.5346, 0.50115677 },
		{ 2001.+138.04167/365.,
		  33.2581, 197.5117, 0.8101696, 86.7103, 355.8796, 0.50103175 },
		{ 2001.+140.04167/365.,
		  33.2541, 197.4414, 0.8101232, 86.7509, 356.7774, 0.50101108 },
		{ 2001.+142.04167/365.,
		  33.2437, 197.3655, 0.8104097, 86.7901, 357.6948, 0.50103624 },
		{ 2001.+144.04167/365.,
		  33.286, 197.1734, 0.8123633, 87.3748, 358.4399, 0.50108211 },
		{ 2001.+146/365.,
		  33.2695, 196.8872, 0.812764, 87.6214, 351.7734, 0.5011419 },
		{ 2001.+148/365.,
		  33.2652, 196.5891, 0.812707, 87.8799, 352.6243, 0.50107993 },
		{ 2001.+150/365.,
		  33.2696, 196.4068, 0.8124701, 88.0215, 353.5125, 0.50099513 },
		{ 2001.+152.04167/365.,
		  33.2764, 196.32, 0.8127915, 88.1444, 1.4555, 0.50088978 },
		{ 2001.+154/365.,
		  33.2706, 196.2277, 0.8132236, 88.2059, 354.7561, 0.50095497 },
		{ 2001.+156/365.,
		  33.2605, 196.068, 0.8138127, 88.3443, 355.5173, 0.5010151 },
		{ 2001.+158.04167/365.,
		  33.2347, 195.8452, 0.8140071, 88.4767, 3.6815, 0.50102788 },
		{ 2001.+160/365.,
		  33.2282, 195.5955, 0.8141851, 88.7159, 356.859, 0.50103202 },
		{ 2001.+162/365.,
		  33.2258, 195.3624, 0.8140935, 88.9385, 357.5611, 0.50100786 },
		{ 2001.+164/365.,
		  33.2321, 195.1705, 0.813813, 89.1009, 358.3401, 0.50096769 },
		{ 2001.+165.95833/365.,
		  33.2323, 195.038, 0.8135843, 89.2126, 351.5967, 0.50092801 },
		{ 2001.+167.95833/365.,
		  33.2288, 194.9375, 0.8135598, 89.2855, 352.4005, 0.5009129 },
		{ 2001.+172/365.,
		  33.1621, 194.5862, 0.8143248, 89.5832, 1.5834, 0.50104566 },
		{ 2001.+173.95833/365.,
		  33.1571, 194.2357, 0.8145014, 89.8958, 354.8376, 0.50106549 },
		{ 2001.+175.95833/365.,
		  33.1646, 193.9294, 0.8142035, 90.1632, 355.6256, 0.50097726 },
		{ 2001.+177.95833/365.,
		  33.174, 193.7619, 0.81391, 90.2919, 356.4529, 0.50090639 },
		{ 2001.+179.91667/365.,
		  33.353, 193.706, 0.8143859, 90.3094, 350.5092, 0.5010413 },
		{ 2001.+181.91667/365.,
		  33.3495, 193.5653, 0.8147865, 90.4253, 351.3761, 0.50110156 },
		{ 2001.+183.91667/365.,
		  33.3452, 193.3558, 0.8152136, 90.6187, 352.1923, 0.50115433 },
		{ 2001.+185.95833/365.,
		  33.2817, 193.0461, 0.8151057, 90.8779, 0.6102, 0.50117709 },
		{ 2001.+187.91667/365.,
		  33.2847, 192.781, 0.8150381, 91.1349, 353.87, 0.50116598 },
		{ 2001.+189.91667/365.,
		  33.2908, 192.5474, 0.8147284, 91.3591, 354.6641, 0.50113018 },
		{ 2001.+191.91667/365.,
		  33.2976, 192.3682, 0.8143169, 91.5245, 355.49, 0.50108577 },
		{ 2001.+193.91667/365.,
		  33.2204, 192.223, 0.8139757, 91.6999, 356.6142, 0.50110617 },
		{ 2001.+195.91667/365.,
		  33.2194, 192.1135, 0.8138859, 91.7825, 357.539, 0.50109597 },
		{ 2001.+197.91667/365.,
		  33.2137, 191.9548, 0.8140888, 91.9094, 358.4688, 0.50114282 },
		{ 2001.+199.91667/365.,
		  33.2346, 191.6851, 0.8143454, 92.1457, 359.3809, 0.50122686 },
		{ 2001.+201.875/365.,
		  33.2437, 191.3264, 0.8142031, 92.4661, 352.7493, 0.50120595 },
		{ 2001.+203.875/365.,
		  33.2623, 191.0588, 0.8136494, 92.6932, 353.6706, 0.50109585 },
		{ 2001.+207.875/365.,
		  33.2597, 190.8278, 0.8133811, 92.8554, 355.5422, 0.50109341 },
		{ 2001.+209.875/365.,
		  33.2608, 190.6695, 0.8136657, 92.9916, 356.4264, 0.5011537 },
		{ 2001.+211.875/365.,
		  33.2646, 190.4491, 0.8138703, 93.1985, 357.2641, 0.50119932 },
		{ 2001.+213.83333/365.,
		  33.2782, 190.2235, 0.8138053, 93.4153, 350.5518, 0.50120399 },
		{ 2001.+215.79167/365.,
		  33.2885, 189.9868, 0.81349, 93.6438, 343.8504, 0.50115375 },
		{ 2001.+217.79167/365.,
		  33.301, 189.7941, 0.8129717, 93.8261, 344.69, 0.50110897 },
		{ 2001.+219.79167/365.,
		  33.3117, 189.6595, 0.8124149, 93.9448, 345.5647, 0.5010746 },
		{ 2001.+221.83333/365.,
		  33.3117, 189.5664, 0.8120032, 94.0185, 353.9901, 0.50108054 },
		{ 2001.+223.83333/365.,
		  33.3122, 189.4758, 0.811873, 94.0801, 354.9209, 0.5010904 },
		{ 2001.+225.83333/365.,
		  33.3105, 189.3166, 0.8119812, 94.2061, 355.848, 0.5011573 },
		{ 2001.+227.83333/365.,
		  33.3142, 189.097, 0.8119617, 94.3732, 356.6876, 0.50119652 },
		{ 2001.+229.79167/365.,
		  33.3362, 188.7704, 0.8115414, 94.6496, 350.0604, 0.50129405 },
		{ 2001.+231.79167/365.,
		  33.3553, 188.5938, 0.8108655, 94.7927, 350.9974, 0.50102232 },
		{ 2001.+233.79167/365.,
		  33.3703, 188.475, 0.8106038, 94.8741, 351.7618, 0.50097392 },
		{ 2001.+235.79167/365.,
		  33.3731, 188.3947, 0.8106907, 94.9203, 352.6613, 0.50105255 },
		{ 2001.+237.79167/365.,
		  33.3761, 188.2467, 0.8109137, 95.0445, 353.5043, 0.5011097 },
		{ 2001.+239.79167/365.,
		  33.3835, 188.0475, 0.8109784, 95.2285, 354.3068, 0.50112998 },
		{ 2001.+241.83333/365.,
		  33.3977, 187.8444, 0.8107589, 95.4137, 2.6163, 0.50113215 },
		{ 2001.+243.79167/365.,
		  33.4114, 187.6632, 0.8103007, 95.5825, 355.8999, 0.50109391 },
		{ 2001.+245.79167/365.,
		  33.4228, 187.5335, 0.8097126, 95.6972, 356.7294, 0.50103701 },
		{ 2001.+247.79167/365.,
		  33.4304, 187.4592, 0.8091605, 95.7492, 357.596, 0.50100925 },
		{ 2001.+249.79167/365.,
		  33.4318, 187.4118, 0.808822, 95.766, 358.4962, 0.50100097 },
		{ 2001.+251.75/365.,
		  33.4284, 187.3446, 0.80878, 95.7988, 351.8924, 0.50103869 },
		{ 2001.+253.75/365.,
		  33.4273, 187.1917, 0.8089075, 95.9149, 352.7907, 0.50110341 },
		{ 2001.+255.79167/365.,
		  33.4389, 186.9424, 0.8088535, 96.1261, 1.1829, 0.50112953 },
		{ 2001.+257.75/365.,
		  33.4602, 186.7015, 0.8083529, 96.3239, 354.5501, 0.50104879 },
		{ 2001.+259.75/365.,
		  33.4779, 186.5897, 0.8077313, 96.3872, 355.4849, 0.50095993 },
		{ 2001.+261.75/365.,
		  33.4852, 186.5644, 0.8077168, 96.4047, 356.1116, 0.5009368 },
		{ 2001.+263.75/365.,
		  33.4839, 186.4988, 0.8079626, 96.4336, 356.9986, 0.50101054 },
		{ 2001.+265.75/365.,
		  33.4853, 186.3591, 0.8082442, 96.5474, 357.8267, 0.5010745 },
		{ 2001.+267.75/365.,
		  33.4922, 186.1794, 0.8083067, 96.7093, 358.6205, 0.5010907 },
		{ 2001.+269.70833/365.,
		  33.5117, 186.0022, 0.8078133, 96.7797, 351.4782, 0.50101141 },
		{ 2001.+271.66667/365.,
		  33.521, 185.8666, 0.8073582, 96.8988, 344.727, 0.50095238 },
		{ 2001.+273.66667/365.,
		  33.528, 185.7841, 0.806868, 96.9595, 345.523, 0.50092121 },
		{ 2001.+275.66667/365.,
		  33.5449, 185.7797, 0.8069732, 97.047, 346.2998, 0.50088676 },
		{ 2001.+277.70833/365.,
		  33.5396, 185.7534, 0.806873, 97.0387, 354.6625, 0.50090225 },
		{ 2001.+280.66667/365.,
		  33.5322, 185.6605, 0.8070493, 97.1012, 168.3476, 0.50096733 },
		{ 2001.+281.66667/365.,
		  33.5322, 185.6423, 0.8070493, 97.1283, 348.6958, 0.50096733 },
		{ 2001.+283.66667/365.,
		  33.5449, 185.2695, 0.807238, 97.4028, 349.6412, 0.50100531 },
		{ 2001.+285.66667/365.,
		  33.5633, 185.0829, 0.8068092, 97.5448, 350.4761, 0.50091295 },
		{ 2001.+287.66667/365.,
		  33.573, 185.0165, 0.8064441, 97.5611, 351.3568, 0.50085247 },
		{ 2001.+289.66667/365.,
		  33.5704, 184.9924, 0.8065661, 97.5385, 352.2399, 0.5008912 },
		{ 2001.+291.66667/365.,
		  33.5388, 184.8755, 0.8068869, 97.5464, 352.8923, 0.50094431 },
		{ 2001.+293.66667/365.,
		  33.5403, 184.7146, 0.8073031, 97.6818, 353.6465, 0.50099794 },
		{ 2001.+295.66667/365.,
		  33.5467, 184.524, 0.8074482, 97.8535, 354.3737, 0.50099894 },
		{ 2001.+297.70833/365.,
		  33.5565, 184.3088, 0.8071538, 97.9963, 2.7024, 0.50100264 },
		{ 2001.+299.66667/365.,
		  33.5638, 184.1817, 0.8068562, 98.1039, 355.9565, 0.50096211 },
		{ 2001.+301.66667/365.,
		  33.5667, 184.1059, 0.8065615, 98.1547, 356.7595, 0.5009184 },
		{ 2001.+303.66667/365.,
		  33.5515, 184.0741, 0.8064109, 98.147, 357.6459, 0.50091108 },
		{ 2001.+305.625/365.,
		  33.5431, 184.021, 0.8065829, 98.1628, 350.9939, 0.50095298 },
		{ 2001.+307.625/365.,
		  33.5362, 183.891, 0.8069895, 98.2524, 351.847, 0.50101105 },
		{ 2001.+309.625/365.,
		  33.5425, 183.6617, 0.8073072, 98.4424, 352.6729, 0.50105462 },
		{ 2001.+311.66667/365.,
		  33.5613, 183.4076, 0.8072395, 98.656, 1.0066, 0.50101432 },
		{ 2001.+313.625/365.,
		  33.5782, 183.2431, 0.8069268, 98.7758, 354.3421, 0.50092565 },
		{ 2001.+315.625/365.,
		  33.5831, 183.1749, 0.8068519, 98.7975, 355.2258, 0.50089105 },
		{ 2001.+317.625/365.,
		  33.5792, 183.1071, 0.8072262, 98.8236, 356.0939, 0.50095963 },
		{ 2001.+319.625/365.,
		  33.5725, 182.9893, 0.8081052, 98.9899, 357.379, 0.50108257 },
		{ 2001.+321.58333/365.,
		  33.58, 182.7775, 0.8085634, 99.1797, 350.6618, 0.50112517 },
		{ 2001.+323.58333/365.,
		  33.5921, 182.5477, 0.8087157, 99.3934, 351.4415, 0.50111095 },
		{ 2001.+325.625/365.,
		  33.6303, 182.3492, 0.8085594, 99.559, 359.7523, 0.50108114 },
		{ 2001.+327.58333/365.,
		  33.6413, 182.2003, 0.8083162, 99.6891, 353.0537, 0.50103802 },
		{ 2001.+329.58333/365.,
		  33.6454, 182.0991, 0.8081206, 99.7665, 353.9051, 0.5010069 },
		{ 2001.+331.58333/365.,
		  33.6402, 182.0249, 0.8081528, 99.8147, 354.7815, 0.50100032 },
		{ 2001.+333.58333/365.,
		  33.6362, 181.9113, 0.8084572, 99.8929, 355.6739, 0.50104269 },
		{ 2001.+335.58333/365.,
		  33.6397, 181.7021, 0.8088975, 100.0642, 356.5495, 0.50112219 },
		{ 2001.+337.58333/365.,
		  33.6603, 181.4035, 0.8091073, 100.325, 357.398, 0.50114009 },
		{ 2001.+339.58333/365.,
		  33.6881, 181.1051, 0.8089119, 100.5703, 358.3966, 0.50121543 },
		{ 2001.+341.54167/365.,
		  33.7073, 180.9641, 0.8085904, 100.6799, 351.8036, 0.50102898 },
		{ 2001.+343.54167/365.,
		  33.7134, 180.8676, 0.8086426, 100.7357, 352.7405, 0.50104654 },
		{ 2001.+345.54167/365.,
		  33.7176, 180.7341, 0.8090547, 100.8243, 353.659, 0.50111662 },
		{ 2001.+347.54167/365.,
		  33.7279, 180.5294, 0.8095702, 101.0048, 354.5114, 0.50117644 },
		{ 2001.+349.54167/365.,
		  33.7474, 180.2661, 0.8098713, 101.2514, 355.3216, 0.50119504 },
		{ 2001.+351.54167/365.,
		  33.7717, 180.0057, 0.8098426, 101.4998, 356.1208, 0.50117814 },
		{ 2001.+353.54167/365.,
		  33.7945, 179.7863, 0.8095657, 101.7069, 356.9367, 0.5011428 },
		{ 2001.+355.54167/365.,
		  33.8124, 179.6232, 0.8092081, 101.8535, 357.7838, 0.50110971 },
		{ 2001.+357.5/365.,
		  33.8308, 179.5155, 0.8090066, 101.9539, 351.0796, 0.50106051 },
		{ 2001.+359.5/365.,
		  33.8345, 179.4041, 0.8089769, 102.0369, 351.9772, 0.50107575 },
		{ 2001.+361.5/365.,
		  33.8251, 179.2308, 0.8092075, 102.1651, 352.809, 0.50110945 },
		{ 2001.+363.5/365.,
		  33.8446, 178.9664, 0.8094741, 102.3943, 353.6801, 0.50117244 },
		{ 2001.+365.5/365.,
		  33.8843, 178.6363, 0.809383, 102.6876, 354.5298, 0.50116072 },
		{ 2002.+2.5/365.,
		  33.9226, 178.3784, 0.8088931, 102.9069, 355.4015, 0.50105666 },
		{ 2002.+4.5416667/365.,
		  33.9499, 178.226, 0.8084785, 103.0136, 3.6037, 0.50095621 },
		{ 2002.+6.5/365.,
		  33.9597, 178.1169, 0.808515, 103.0853, 356.9601, 0.5009993 },
		{ 2002.+8.5/365.,
		  33.9744, 177.9523, 0.8088399, 103.2258, 357.735, 0.50104915 },
		{ 2002.+10.458333/365.,
		  33.997, 177.7215, 0.8090994, 103.4373, 350.9841, 0.50110616 },
		{ 2002.+12.458333/365.,
		  34.0273, 177.462, 0.8090932, 103.6847, 351.7192, 0.50110276 },
		{ 2002.+14.458333/365.,
		  34.0603, 177.2203, 0.8087749, 103.9165, 352.4537, 0.50107511 },
		{ 2002.+16.5/365.,
		  34.1042, 177.0075, 0.808238, 104.1101, 0.6661, 0.50102247 },
		{ 2002.+18.458333/365.,
		  34.123, 176.875, 0.8077019, 104.2267, 353.9347, 0.50097724 },
		{ 2002.+20.458333/365.,
		  34.1345, 176.7775, 0.8073131, 104.3011, 354.7513, 0.50096327 },
		{ 2002.+22.458333/365.,
		  34.1409, 176.6701, 0.8071674, 104.379, 355.5884, 0.50099262 },
		{ 2002.+24.458333/365.,
		  34.1508, 176.5061, 0.8071896, 104.5068, 356.3825, 0.50104122 },
		{ 2002.+26.416667/365.,
		  34.1811, 176.2436, 0.8071968, 104.7341, 349.6761, 0.50108968 },
		{ 2002.+28.416667/365.,
		  34.2296, 175.9468, 0.8067823, 104.9928, 350.475, 0.50103493 },
		{ 2002.+30.458333/365.,
		  34.2643, 175.7517, 0.8060596, 105.1487, 358.8256, 0.5009512 },
		{ 2002.+32.416667/365.,
		  34.2804, 175.666, 0.8056129, 105.1906, 352.1854, 0.50090591 },
		{ 2002.+34.416667/365.,
		  34.2891, 175.5734, 0.8056237, 105.2451, 353.0368, 0.50097958 },
		{ 2002.+36.458333/365.,
		  34.3069, 175.4114, 0.8058025, 105.3808, 1.3056, 0.50103194 },
		{ 2002.+38.416667/365.,
		  34.3333, 175.2023, 0.8058292, 105.5718, 354.5345, 0.50106458 },
		{ 2002.+40.416667/365.,
		  34.3662, 174.9865, 0.80556, 105.7752, 355.2605, 0.50105844 },
		{ 2002.+42.416667/365.,
		  34.3961, 174.8042, 0.8050256, 105.9464, 355.9957, 0.50101809 },
		{ 2002.+44.416667/365.,
		  34.4237, 174.6678, 0.8039598, 105.9613, 356.6731, 0.5009508 },
		{ 2002.+46.375/365.,
		  34.4353, 174.5933, 0.8033518, 106.0148, 349.9539, 0.50092307 },
		{ 2002.+48.375/365.,
		  34.439, 174.5406, 0.8029533, 106.0394, 350.7786, 0.50092502 },
		{ 2002.+50.375/365.,
		  34.4407, 174.464, 0.8028119, 106.0831, 351.6165, 0.50096476 },
		{ 2002.+52.416667/365.,
		  34.4402, 174.3254, 0.8031277, 106.2788, 359.824, 0.50100846 },
		{ 2002.+54.375/365.,
		  34.4727, 174.0987, 0.8029815, 106.4686, 353.108, 0.50103373 },
		{ 2002.+56.375/365.,
		  34.5163, 173.8708, 0.8023984, 106.6556, 353.9076, 0.5009691 },
		{ 2002.+58.416667/365.,
		  34.5583, 173.7441, 0.8017078, 106.7446, 2.179, 0.50086842 },
		{ 2002.+60.375/365.,
		  34.5621, 173.711, 0.8013787, 106.7283, 355.5348, 0.50084771 },
		{ 2002.+62.375/365.,
		  34.5648, 173.6397, 0.801504, 106.7581, 356.3661, 0.50093432 },
		{ 2002.+64.375/365.,
		  34.5783, 173.4962, 0.8016934, 106.8717, 357.1367, 0.50099641 },
		{ 2002.+66.333333/365.,
		  34.6044, 173.314, 0.8016631, 107.0304, 350.4759, 0.50102796 },
		{ 2002.+68.333333/365.,
		  34.6297, 173.1479, 0.8013236, 107.1813, 351.2154, 0.50100245 },
		{ 2002.+70.333333/365.,
		  34.652, 173.0273, 0.8007486, 107.2828, 351.9714, 0.50104595 },
		{ 2002.+72.375/365.,
		  34.6627, 172.9524, 0.8000688, 107.3264, 0.2797, 0.50092519 },
		{ 2002.+74.333333/365.,
		  34.6623, 172.9235, 0.7995981, 107.3273, 353.5915, 0.50091174 },
		{ 2002.+76.333333/365.,
		  34.6556, 172.8983, 0.7993854, 107.3179, 354.4418, 0.50092247 },
		{ 2002.+78.333333/365.,
		  34.6523, 172.8308, 0.7994169, 107.348, 355.2966, 0.50096877 },
		{ 2002.+80.375/365.,
		  34.6613, 172.6861, 0.7995138, 107.4591, 3.715, 0.50103982 },
		{ 2002.+82.333333/365.,
		  34.6921, 172.4853, 0.7993622, 107.6204, 357.0331, 0.50104369 },
		{ 2002.+84.333333/365.,
		  34.7262, 172.3158, 0.7988255, 107.7452, 357.8785, 0.500951 },
		{ 2002.+86.333333/365.,
		  34.7452, 172.2499, 0.7982643, 107.7586, 358.8215, 0.50088597 },
		{ 2002.+88.291667/365.,
		  34.7397, 172.2338, 0.7982139, 107.7217, 352.2311, 0.50091115 },
		{ 2002.+90.291667/365.,
		  34.7392, 172.1552, 0.7985647, 107.7566, 353.1045, 0.50101165 },
		{ 2002.+92.291667/365.,
		  34.753, 172.0176, 0.7990907, 107.9332, 353.8855, 0.50105163 },
		{ 2002.+94.291667/365.,
		  34.7751, 171.8448, 0.7991076, 108.0845, 354.6591, 0.50105317 },
		{ 2002.+96.291667/365.,
		  34.7959, 171.6979, 0.798835, 108.2134, 355.4341, 0.50102498 },
		{ 2002.+98.291667/365.,
		  34.8096, 171.5999, 0.798411, 108.2907, 356.2322, 0.5009906 },
		{ 2002.+100.29167/365.,
		  34.8056, 171.5412, 0.7977734, 108.2705, 357.208, 0.50097631 },
		{ 2002.+102.25/365.,
		  34.7986, 171.5174, 0.7975583, 108.2611, 350.5815, 0.50097788 },
		{ 2002.+104.25/365.,
		  34.7888, 171.478, 0.7976199, 108.2623, 351.4867, 0.50100207 },
		{ 2002.+106.25/365.,
		  34.7881, 171.3786, 0.7978749, 108.3222, 352.3865, 0.50106175 },
		{ 2002.+108.25/365.,
		  34.8059, 171.2031, 0.79808, 108.4588, 353.2649, 0.50110365 },
		{ 2002.+110.25/365.,
		  34.8339, 170.9949, 0.797911, 108.594, 354.3538, 0.50112494 },
		{ 2002.+112.25/365.,
		  34.8621, 170.8521, 0.7974965, 108.6912, 355.278, 0.50102992 },
		{ 2002.+114.25/365.,
		  34.8676, 170.8007, 0.7972308, 108.6903, 356.2459, 0.50098266 },
		{ 2002.+116.20833/365.,
		  34.8459, 170.7475, 0.7974781, 108.6918, 349.7833, 0.50107555 },
		{ 2002.+118.20833/365.,
		  34.8503, 170.6265, 0.7980202, 108.7713, 350.7139, 0.50115979 },
		{ 2002.+120.20833/365.,
		  34.8714, 170.4385, 0.7984235, 108.9315, 351.5848, 0.50119172 },
		{ 2002.+122.20833/365.,
		  34.908, 170.2484, 0.798447, 109.0966, 352.4661, 0.5011901 },
		{ 2002.+124.20833/365.,
		  34.9316, 170.09, 0.7982296, 109.2373, 353.3309, 0.50115274 },
		{ 2002.+126.20833/365.,
		  34.9462, 169.984, 0.7979139, 109.322, 354.2232, 0.50111675 },
		{ 2002.+128.20833/365.,
		  34.9492, 169.9193, 0.7976844, 109.36, 355.1478, 0.50108646 },
		{ 2002.+134.20833/365.,
		  34.9531, 169.6189, 0.7985486, 109.6369, 358.0524, 0.50119902 },
		{ 2002.+136.16667/365.,
		  34.9868, 169.3937, 0.798731, 109.8251, 351.4611, 0.50121993 },
		{ 2002.+138.16667/365.,
		  35.0294, 169.1746, 0.7985377, 110.0061, 352.3841, 0.50116404 },
		{ 2002.+140.16667/365.,
		  35.0582, 169.0396, 0.7981815, 110.0968, 353.3394, 0.50107972 },
		{ 2002.+142.16667/365.,
		  35.0648, 168.9702, 0.7981224, 110.1189, 354.3263, 0.50105933 },
		{ 2002.+144.16667/365.,
		  35.0689, 168.8756, 0.7985128, 110.1701, 355.2957, 0.50113766 },
		{ 2002.+146.16667/365.,
		  35.0876, 168.7008, 0.7990627, 110.3111, 356.2061, 0.50120898 },
		{ 2002.+148.16667/365.,
		  35.1232, 168.4733, 0.7993808, 110.5169, 357.0636, 0.50123401 },
		{ 2002.+150.16667/365.,
		  35.1528, 168.2611, 0.7990035, 110.6129, 358.53, 0.501262 },
		{ 2002.+152.125/365.,
		  35.1846, 168.0885, 0.7987131, 110.77, 351.905, 0.50121293 },
		{ 2002.+154.125/365.,
		  35.2045, 167.9689, 0.7983758, 110.8705, 352.83, 0.50117031 },
		{ 2002.+156.56591/365.,
		  35.2045, 167.9286, 0.7983758, 110.9282, 73.223, 0.50117031 },
		{ 2002.+158.16667/365.,
		  35.2311, 167.7964, 0.7981855, 110.9803, 2.2786, 0.50117597 },
		{ 2002.+160.125/365.,
		  35.2422, 167.6631, 0.7984329, 111.0791, 355.7403, 0.50123187 },
		{ 2002.+162.125/365.,
		  35.2725, 167.4513, 0.7987017, 111.2562, 356.7006, 0.50128103 },
		{ 2002.+164.125/365.,
		  35.3255, 167.1933, 0.7986731, 111.4803, 357.6385, 0.50127263 },
		{ 2002.+166.08333/365.,
		  35.4235, 167.011, 0.7982515, 111.6364, 350.6481, 0.50109285 },
		{ 2002.+168.08333/365.,
		  35.4525, 166.8907, 0.7978562, 111.7141, 351.5703, 0.50103037 },
		{ 2002.+170.08333/365.,
		  35.4638, 166.8039, 0.7978593, 111.7581, 352.5067, 0.50105562 },
		{ 2002.+172.125/365.,
		  35.437, 166.6627, 0.7983689, 111.9033, 1.0177, 0.50114517 },
		{ 2002.+174.08333/365.,
		  35.4709, 166.4556, 0.7987348, 112.0839, 354.3587, 0.50121155 },
		{ 2002.+176.08333/365.,
		  35.5193, 166.2165, 0.7988072, 112.3071, 355.1757, 0.50121888 },
		{ 2002.+178.08333/365.,
		  35.5661, 166.0012, 0.798532, 112.5109, 355.9879, 0.50117706 },
		{ 2002.+180.08333/365.,
		  35.6032, 165.84, 0.7980547, 112.6587, 356.8236, 0.50113091 },
		{ 2002.+182.04167/365.,
		  35.6259, 165.7287, 0.7975856, 112.752, 350.1769, 0.50108828 },
		{ 2002.+184.04167/365.,
		  35.638, 165.6449, 0.7972877, 112.8112, 351.0768, 0.50108901 },
		{ 2002.+186.04167/365.,
		  35.6449, 165.5433, 0.7971095, 112.8466, 352.1229, 0.50114181 },
		{ 2002.+188.04167/365.,
		  35.6673, 165.3883, 0.7972282, 112.9693, 353.0519, 0.50119713 },
		{ 2002.+190.04167/365.,
		  35.7127, 165.1601, 0.7972678, 113.1651, 353.9591, 0.50123607 },
		{ 2002.+192.08333/365.,
		  35.7726, 164.905, 0.7969281, 113.4037, 2.5429, 0.50124497 },
		{ 2002.+194.04167/365.,
		  35.8235, 164.7287, 0.7962704, 113.5414, 355.9773, 0.50115718 },
		{ 2002.+196.04167/365.,
		  35.8459, 164.6412, 0.7958133, 113.5856, 356.9702, 0.50110931 },
		{ 2002.+198.04167/365.,
		  35.8582, 164.5538, 0.7958071, 113.6322, 357.9595, 0.50116064 },
		{ 2002.+200.04167/365.,
		  35.8835, 164.4049, 0.796043, 113.7491, 358.9012, 0.50122892 },
		{ 2002.+202/365.,
		  35.9269, 164.2081, 0.7961766, 113.9064, 352.2753, 0.50127532 },
		{ 2002.+204/365.,
		  35.9778, 163.9975, 0.795962, 114.1038, 353.1362, 0.50126423 },
		{ 2002.+206.04167/365.,
		  36.0241, 163.8182, 0.7954233, 114.2724, 1.5212, 0.50124899 },
		{ 2002.+208/365.,
		  36.055, 163.6981, 0.7947738, 114.3782, 354.8985, 0.50117884 },
		{ 2002.+210/365.,
		  36.0707, 163.6229, 0.7941906, 114.4333, 355.8297, 0.50114944 },
		{ 2002.+212/365.,
		  36.0648, 163.5695, 0.7938138, 114.4655, 356.8644, 0.50117012 },
		{ 2002.+213.95833/365.,
		  36.0739, 163.4832, 0.793685, 114.5205, 350.3276, 0.50120386 },
		{ 2002.+215.95833/365.,
		  36.0996, 163.3364, 0.7936705, 114.6341, 351.2999, 0.50125554 },
		{ 2002.+217.79167/365.,
		  36.1432, 163.1343, 0.7934978, 114.8039, 322.1755, 0.50125524 },
		{ 2002.+219.95833/365.,
		  36.2222, 162.9242, 0.7929418, 114.9683, 353.2053, 0.50121977 },
		{ 2002.+221.95833/365.,
		  36.259, 162.8126, 0.7921522, 115.0362, 354.1903, 0.50111905 },
		{ 2002.+223.95833/365.,
		  36.2678, 162.7668, 0.7917214, 115.034, 355.2112, 0.50110529 },
		{ 2002.+225.95833/365.,
		  36.2762, 162.6915, 0.7917547, 115.0664, 356.2094, 0.50117484 },
		{ 2002.+226/365.,
		  36.2762, 162.6909, 0.7917547, 115.0673, 3.727, 0.50117484 },
		{ 2002.+227.95833/365.,
		  36.3023, 162.5513, 0.7918973, 115.1757, 357.1506, 0.50124408 },
		{ 2002.+229.95833/365.,
		  36.343, 162.3756, 0.7918153, 115.3314, 358.0458, 0.5012718 },
		{ 2002.+231.95833/365.,
		  36.3856, 162.2111, 0.7914026, 115.4818, 358.9263, 0.50125796 },
		{ 2002.+233.95833/365.,
		  36.4181, 162.0909, 0.7907504, 115.5876, 359.8219, 0.50121982 },
		{ 2002.+235.95833/365.,
		  36.4411, 162.0513, 0.7900527, 115.5805, 0.1912, 0.5010856 },
		{ 2002.+237.91667/365.,
		  36.4433, 162.018, 0.7894957, 115.5866, 353.5918, 0.5010712 },
		{ 2002.+239.91667/365.,
		  36.4397, 161.9841, 0.789188, 115.5874, 354.5317, 0.50108714 },
		{ 2002.+241.91667/365.,
		  36.443, 161.9128, 0.7891013, 115.6225, 355.4764, 0.50112957 },
		{ 2002.+243.91667/365.,
		  36.4667, 161.782, 0.78906, 115.7174, 356.4076, 0.50117178 },
		{ 2002.+245.91667/365.,
		  36.5114, 161.6065, 0.7887805, 115.8565, 357.3194, 0.50116667 },
		{ 2002.+247.91667/365.,
		  36.5581, 161.4597, 0.7881077, 115.9615, 358.2366, 0.50108745 },
		{ 2002.+249.91667/365.,
		  36.5759, 161.4038, 0.7873853, 115.966, 359.2006, 0.50100535 },
		{ 2002.+251.875/365.,
		  36.5712, 161.3863, 0.787146, 115.9296, 352.6774, 0.5010259 },
		{ 2002.+253.875/365.,
		  36.575, 161.3145, 0.7873306, 115.954, 353.6295, 0.50111886 },
		{ 2002.+255.875/365.,
		  36.5851, 161.1637, 0.7875265, 116.0755, 354.416, 0.50116198 },
		{ 2002.+257.875/365.,
		  36.6196, 161.0108, 0.7873996, 116.2061, 355.2631, 0.50116846 },
		{ 2002.+259.875/365.,
		  36.6503, 160.8833, 0.7870038, 116.3299, 356.1097, 0.50112999 },
		{ 2002.+261.875/365.,
		  36.6683, 160.8049, 0.7864068, 116.3875, 356.9773, 0.50109456 },
		{ 2002.+263.875/365.,
		  36.6745, 160.7772, 0.7859368, 116.3816, 357.9954, 0.50107212 },
		{ 2002.+265.83333/365.,
		  36.6651, 160.7642, 0.7855771, 116.3607, 351.4352, 0.5010714 },
		{ 2002.+267.83333/365.,
		  36.6545, 160.735, 0.7854795, 116.3501, 352.4074, 0.50110504 },
		{ 2002.+269.83333/365.,
		  36.6436, 160.6378, 0.785546, 116.4001, 353.4167, 0.50125714 },
		{ 2002.+271.83333/365.,
		  36.6669, 160.5035, 0.7856165, 116.4917, 354.3775, 0.50120051 },
		{ 2002.+273.83333/365.,
		  36.7081, 160.346, 0.785363, 116.609, 355.3212, 0.50117974 },
		{ 2002.+275.83333/365.,
		  36.742, 160.2353, 0.7847962, 116.6726, 356.2814, 0.50108145 },
		{ 2002.+277.79167/365.,
		  36.75, 160.2066, 0.7842464, 116.6228, 349.8723, 0.50103614 },
		{ 2002.+279.79167/365.,
		  36.7404, 160.1829, 0.7843116, 116.588, 350.9076, 0.50108879 },
		{ 2002.+281.79167/365.,
		  36.7467, 160.0888, 0.784715, 116.6334, 351.8901, 0.50119586 },
		{ 2002.+283.83333/365.,
		  36.7913, 159.9639, 0.7851761, 116.7953, 0.4211, 0.50125069 },
		{ 2002.+285.79167/365.,
		  36.825, 159.8094, 0.785105, 116.9256, 353.8095, 0.50122852 },
		{ 2002.+287.79167/365.,
		  36.8515, 159.6928, 0.7847543, 117.0206, 354.7244, 0.50118969 },
		{ 2002.+289.79167/365.,
		  36.863, 159.6243, 0.7843176, 117.0634, 355.6668, 0.50115666 },
		{ 2002.+291.79167/365.,
		  36.856, 159.5724, 0.7839038, 117.0355, 356.6756, 0.50117104 },
		{ 2002.+293.79167/365.,
		  36.8453, 159.5457, 0.7838313, 117.0219, 357.6819, 0.50114021 },
		{ 2002.+295.79167/365.,
		  36.8394, 159.4872, 0.7839847, 117.0383, 358.696, 0.50118331 },
		{ 2002.+297.79167/365.,
		  36.8521, 159.3549, 0.7842766, 117.1258, 359.8312, 0.50127516 },
		{ 2002.+299.75/365.,
		  36.8865, 159.1927, 0.7843913, 117.2482, 353.3238, 0.50129813 },
		{ 2002.+301.75/365.,
		  36.9316, 159.0309, 0.7841811, 117.3691, 354.327, 0.50124548 },
		{ 2002.+303.75/365.,
		  36.9601, 158.9286, 0.7837589, 117.4235, 355.3553, 0.50115915 },
		{ 2002.+305.75/365.,
		  36.963, 158.8751, 0.7835592, 117.3922, 356.4996, 0.50114266 },
		{ 2002.+307.66667/365.,
		  36.9606, 158.8097, 0.7838544, 117.3984, 342.5413, 0.50128075 },
		{ 2002.+309.70833/365.,
		  36.9822, 158.6742, 0.7844277, 117.493, 351.0811, 0.50130989 },
		{ 2002.+311.75/365.,
		  37.026, 158.5012, 0.7847952, 117.6674, 359.6789, 0.50136245 },
		{ 2002.+313.70833/365.,
		  37.0698, 158.324, 0.7847205, 117.8223, 353.1201, 0.50133539 },
		{ 2002.+315.70833/365.,
		  37.1023, 158.1942, 0.7844062, 117.9306, 354.0984, 0.50128454 },
		{ 2002.+317.70833/365.,
		  37.1178, 158.1111, 0.7840716, 117.9878, 355.1074, 0.50124758 },
		{ 2002.+319.70833/365.,
		  37.12, 158.0442, 0.7838853, 118.0131, 356.2099, 0.50125505 },
		{ 2002.+321.70833/365.,
		  37.1197, 157.9783, 0.7839555, 118.0405, 357.2779, 0.50127307 },
		{ 2002.+323.70833/365.,
		  37.131, 157.8714, 0.7842271, 118.1079, 358.3461, 0.5013179 },
		{ 2002.+325.70833/365.,
		  37.1651, 157.7016, 0.7845006, 118.2388, 359.3944, 0.50136688 },
		{ 2002.+327.66667/365.,
		  37.219, 157.5035, 0.784529, 118.4002, 352.9012, 0.50136393 },
		{ 2002.+329.66667/365.,
		  37.2737, 157.3325, 0.7842192, 118.5322, 353.9292, 0.50130157 },
		{ 2002.+333.70833/365.,
		  37.3029, 157.1545, 0.7838484, 118.6136, 3.6763, 0.50124364 },
		{ 2002.+335.66667/365.,
		  37.3191, 157.046, 0.7842419, 118.6745, 357.2444, 0.50131366 },
		{ 2002.+337.66667/365.,
		  37.3599, 156.864, 0.7847087, 118.8202, 358.2698, 0.50139724 },
		{ 2002.+339.66667/365.,
		  37.4194, 156.6474, 0.784863, 119.0134, 359.2426, 0.50141367 },
		{ 2002.+341.625/365.,
		  37.4694, 156.4531, 0.7845782, 119.1853, 352.6077, 0.5013451 },
		{ 2002.+343.58333/365.,
		  37.5099, 156.3155, 0.7841379, 119.3044, 346.0623, 0.50128851 },
		{ 2002.+345.58333/365.,
		  37.5334, 156.2231, 0.7837353, 119.3724, 347.0701, 0.50125948 },
		{ 2002.+347.58333/365.,
		  37.5456, 156.1468, 0.7835254, 119.4181, 348.1047, 0.50126947 },
		{ 2002.+349.625/365.,
		  37.5399, 156.0152, 0.7837312, 119.5186, 356.589, 0.50129755 },
		{ 2002.+351.58333/365.,
		  37.5692, 155.8732, 0.7839016, 119.6246, 350.107, 0.50135121 },
		{ 2002.+353.58333/365.,
		  37.6482, 155.737, 0.783869, 119.7694, 350.9238, 0.50133328 },
		{ 2002.+355.58333/365.,
		  37.7192, 155.5304, 0.783619, 119.9428, 351.8902, 0.50129479 },
		{ 2002.+357.58333/365.,
		  37.7949, 155.3544, 0.7831032, 120.0482, 352.8155, 0.50119915 },
		{ 2002.+359.58333/365.,
		  37.8246, 155.265, 0.7826459, 120.0929, 353.8223, 0.50114741 },
		{ 2002.+361.58333/365.,
		  37.8409, 155.1833, 0.7825946, 120.128, 354.8415, 0.50118952 },
		{ 2002.+363.58333/365.,
		  37.8715, 155.0496, 0.7828634, 120.2213, 355.8263, 0.50126157 },
		{ 2002.+365.58333/365.,
		  37.9272, 154.8465, 0.7830092, 120.3783, 356.6889, 0.50130657 },
		{ 2003.+2.5416667/365.,
		  37.9937, 154.6407, 0.7828571, 120.5663, 350.047, 0.50129019 },
		{ 2003.+4.5416667/365.,
		  38.0534, 154.4714, 0.7823396, 120.7214, 350.9263, 0.50123353 },
		{ 2003.+6.5416667/365.,
		  38.0924, 154.3572, 0.7816787, 120.8188, 351.8359, 0.50117956 },
		{ 2003.+8.5416667/365.,
		  38.1133, 154.2833, 0.7811102, 120.868, 352.781, 0.50117377 },
		{ 2003.+10.541667/365.,
		  38.1282, 154.2168, 0.7808314, 120.9092, 353.8697, 0.5012063 },
		{ 2003.+12.541667/365.,
		  38.1457, 154.1236, 0.780722, 120.9683, 354.8681, 0.50124896 },
		{ 2003.+14.541667/365.,
		  38.183, 153.9827, 0.7806553, 121.0694, 355.9046, 0.50131451 },
		{ 2003.+16.541667/365.,
		  38.2456, 153.8008, 0.7804552, 121.2191, 356.8831, 0.50132935 },
		{ 2003.+18.541667/365.,
		  38.3148, 153.6266, 0.7798783, 121.3596, 357.8514, 0.50127931 },
		{ 2003.+20.541667/365.,
		  38.3611, 153.5197, 0.7790976, 121.425, 358.8469, 0.5011912 },
		{ 2003.+22.541667/365.,
		  38.3767, 153.4649, 0.7785788, 121.4323, 359.8802, 0.50116505 },
		{ 2003.+26.541667/365.,
		  38.4238, 153.2622, 0.7786333, 121.5581, 1.7833, 0.5012662 },
		{ 2003.+30.5/365.,
		  38.5357, 152.9342, 0.7781365, 121.8454, 356.0634, 0.50127082 },
		{ 2003.+34.5/365.,
		  38.6018, 152.7419, 0.7766385, 122.0079, 357.842, 0.50114736 },
		{ 2003.+36.458333/365.,
		  38.6074, 152.7055, 0.7760152, 122.0153, 351.2895, 0.50114422 },
		{ 2003.+38.458333/365.,
		  38.6083, 152.6652, 0.7756558, 122.0198, 352.2763, 0.5011688 },
		{ 2003.+40.416667/365.,
		  38.6219, 152.5987, 0.7754358, 122.0347, 345.7997, 0.50121674 },
		{ 2003.+42.416667/365.,
		  38.654, 152.4788, 0.7753019, 122.1184, 346.7806, 0.50125721 },
		{ 2003.+44.416667/365.,
		  38.7084, 152.331, 0.7749393, 122.231, 347.7425, 0.50124924 },
		{ 2003.+46.5/365.,
		  38.7622, 152.2076, 0.7741986, 122.3152, 3.7411, 0.50120818 },
		{ 2003.+48.458333/365.,
		  38.7844, 152.1537, 0.7733956, 122.3198, 357.2235, 0.50110958 },
		{ 2003.+50.458333/365.,
		  38.7817, 152.1323, 0.7729984, 122.2861, 358.2583, 0.5011034 },
		{ 2003.+52.458333/365.,
		  38.7875, 152.0719, 0.7730387, 122.2964, 359.2706, 0.50118245 },
		{ 2003.+54.458333/365.,
		  38.8165, 151.9547, 0.7731342, 122.376, 0.2259, 0.50124916 },
		{ 2003.+58.458333/365.,
		  38.9018, 151.7012, 0.7724279, 122.5848, 2.0366, 0.50124172 },
		{ 2003.+60.416667/365.,
		  38.9252, 151.6307, 0.7717125, 122.6334, 355.4399, 0.50118515 },
		{ 2003.+62.416667/365.,
		  38.9279, 151.6019, 0.7710333, 122.6331, 356.3981, 0.50114032 },
		{ 2003.+64.458333/365.,
		  38.9136, 151.5776, 0.770555, 122.6242, 1.7949, 0.50122682 },
		{ 2003.+68.416667/365.,
		  38.9093, 151.4841, 0.770332, 122.6339, 356.4219, 0.50130145 },
		{ 2003.+70.416667/365.,
		  38.9369, 151.375, 0.7702522, 122.703, 357.4746, 0.50133439 },
		{ 2003.+72.416667/365.,
		  38.9924, 151.2579, 0.7698566, 122.7757, 358.5467, 0.50130516 },
		{ 2003.+74.375/365.,
		  39.0271, 151.1717, 0.7691954, 122.8165, 352.0789, 0.50122309 },
		{ 2003.+76.375/365.,
		  39.0311, 151.1476, 0.7685902, 122.7824, 353.1747, 0.50116576 },
		{ 2003.+78.375/365.,
		  39.0182, 151.1321, 0.7684848, 122.7357, 354.2948, 0.50119554 },
		{ 2003.+80.375/365.,
		  39.0235, 151.0603, 0.7687615, 122.7538, 355.375, 0.50128857 },
		{ 2003.+82.375/365.,
		  39.0542, 150.9358, 0.7689591, 122.8398, 356.3923, 0.50134321 },
		{ 2003.+84.375/365.,
		  39.0936, 150.8065, 0.7688088, 122.9423, 357.3746, 0.50133476 },
		{ 2003.+86.375/365.,
		  39.1206, 150.7082, 0.7682894, 123.0069, 358.2723, 0.50126708 },
		{ 2003.+88.333333/365.,
		  39.131, 150.6577, 0.7677422, 123.0297, 351.7531, 0.50120773 },
		{ 2003.+90.333333/365.,
		  39.1241, 150.6383, 0.7673056, 123.0131, 352.7869, 0.50119173 },
		{ 2003.+92.333333/365.,
		  39.1093, 150.6219, 0.7671204, 122.9864, 353.8467, 0.50121178 },
		{ 2003.+94.375/365.,
		  39.0978, 150.5808, 0.7673887, 123.042, 2.3538, 0.50123516 },
		{ 2003.+96.333333/365.,
		  39.1093, 150.4946, 0.7675478, 123.0842, 355.8802, 0.50128004 },
		{ 2003.+98.333333/365.,
		  39.1435, 150.3733, 0.7675758, 123.1647, 356.904, 0.50129789 },
		{ 2003.+100.33333/365.,
		  39.1966, 150.2563, 0.7669069, 123.1382, 357.8008, 0.50122559 },
		{ 2003.+102.29167/365.,
		  39.2239, 150.1817, 0.766414, 123.1631, 351.2948, 0.50114326 },
		{ 2003.+104.29167/365.,
		  39.2223, 150.1547, 0.7661029, 123.1299, 352.3492, 0.50110086 },
		{ 2003.+106.33333/365.,
		  39.2045, 150.1131, 0.7665419, 123.1759, 0.9391, 0.5011551 },
		{ 2003.+108.29167/365.,
		  39.2223, 150.008, 0.7670124, 123.2292, 354.4399, 0.5012664 },
		{ 2003.+110.29167/365.,
		  39.265, 149.8562, 0.7672891, 123.3453, 355.3944, 0.5013002 },
		{ 2003.+112.29167/365.,
		  39.3109, 149.7142, 0.7671639, 123.4618, 356.3236, 0.50127036 },
		{ 2003.+116.25/365.,
		  39.3577, 149.5698, 0.7663477, 123.5441, 350.5088, 0.50115863 },
		{ 2003.+118.25/365.,
		  39.3539, 149.5344, 0.7661285, 123.5412, 351.4973, 0.50114905 },
	};

	unsigned nPoints = NUM_OF(orbitalElem);
	for (unsigned n=0; n<nPoints; ++n) {
		//	convert time to MET
		orbitalElem[n][0] = STime(orbitalElem[n][0], STime::DYN) -
								Xmm::missionReferenceTime;
		//	convert degrees to rad
		orbitalElem[n][1] *= DEG_TO_RAD;
		orbitalElem[n][2] *= DEG_TO_RAD;
		orbitalElem[n][4] *= DEG_TO_RAD;
		orbitalElem[n][5] *= DEG_TO_RAD;

		//	convert orbital period [rev/d] -> [s/rev]
		orbitalElem[n][6] = 86400./orbitalElem[n][6];
	}

	double t = time - Xmm::missionReferenceTime;	//	time at which
													//	orbital elements
													//	are sought
	double parValues[N_PAR];
	for (unsigned n=0; n<N_PAR; ++n) {
		std::vector<double> x(nPoints),
					   y(nPoints);
		for (unsigned i=0; i<nPoints; ++i) {
			x[i] = orbitalElem[i][0];
			y[i] = orbitalElem[i][n+1];
		}
		Slatec::CubicHermiteInterpolant<double> f(x, y);
		parValues[n] = f(t);
		//cout<<n<<' '<<parValues[n]<<endl;
	}

	OrbitalElements orb = { parValues[5],	//	period
							114000.,		//	apogee [km]
							7000.,			//	perigee [km]
							parValues[2],	//	eccentricity
							parValues[0],	//	orbit inclination [rad]
							parValues[1],	//	R.A. of ascending node [rad]
							parValues[3],	//	argument of perigee [rad]
						  };

	orbitalElements(orb,	//	orbital elements valid at <time>

							//	set time of perigee pass in this revolution
							//	to be time in MET <t> minus mean anomaly
							//	at <t> over 2Pi * orbital period at <t>
					t-parValues[4]/M_TWO_PI*parValues[5]);
} // XmmOrbitSimulator::XmmOrbitSimulator()


XmmGroundStation::XmmGroundStation(const EarthObject &where,
								   unsigned short id,
								   double timingDelay)
	: EarthPositionLocator(where, Xmm::missionReferenceTime),
	  _id(id), _timingDelay(timingDelay)
{
} // XmmGroundStation::XmmGroundStation()

//
//	-------------------------( XMM )------------------------------------------
//
const XmmPayLoad &XmmMission::payLoad()
{
	static XmmPayLoad *xmmPayLoadObj = new XmmPayLoad;

	return *xmmPayLoadObj;
} // XmmMission::payLoad()


const XmmDataHandler &XmmMission::dataHandler()
{
	static XmmDataHandler *xmmDataHandler = new XmmDataHandler;

	return *xmmDataHandler;
} // XmmMission::dataHandler()


const XmmOrbitSimulator &XmmMission::orbitSimulator()
{
	static XmmOrbitSimulator *xmmOrbitSimulator = new XmmOrbitSimulator;

	return *xmmOrbitSimulator;
} // XmmMission::orbitSimulator()


XmmMission::GroundStations XmmMission::_groundStations;


void XmmMission::checkGSset()
{
	if (!_groundStations.size()) {
		//
		//	following data taken from ESOC file Station.dat and data provided
		//	by Ed Serpell on 2001-02-28; the timing delay numbers are
		//	used as of XMCS v72.
		//
		XmmGroundStation
			Vilspa1(EarthObject(356.04841574 * DEG_TO_RAD,
								40.44256248 * DEG_TO_RAD,EarthObject::GEODETIC,
								656.455,
								"Villafranca-1"), 5, 474.17e-6),
												//	Vilspa2 value in absence
												//	of more specific data
			Vilspa2(EarthObject(356.04740072 * DEG_TO_RAD,
								40.44559434 * DEG_TO_RAD,EarthObject::GEODETIC,
								664.432,
								"Villafranca-2"), 13, 474.17e-6),
			Vilspa3(EarthObject(356.04841847 * DEG_TO_RAD,
								40.44333913 * DEG_TO_RAD,EarthObject::GEODETIC,
								650.833,
								"Villafranca-3"), 14, 474.17e-6),
												//	Vilspa2 value in absence
												//	of more specific data
			Perth(	EarthObject(115.88515805 * DEG_TO_RAD,
								-31.80252200 *DEG_TO_RAD,EarthObject::GEODETIC,
								22.160,
								"Perth"), 22, 465.54e-6),
			Kourou( EarthObject(307.19533724 * DEG_TO_RAD,
								5.25143810 * DEG_TO_RAD, EarthObject::GEODETIC,
								-14.558,
								"Kourou"), 21, 474.87e-6),
			Santiago(EarthObject(289.33363100 * DEG_TO_RAD,
								-33.15099900 *DEG_TO_RAD,EarthObject::GEODETIC,
								729.0,
								"Santiago"), 41, 597.00e-6 + 110.e-6),
													//	additional 110us yield
													//	better data alignment
			NewNorcia(EarthObject(116.19150227 * DEG_TO_RAD,
								  -31.04822306*DEG_TO_RAD,
								  EarthObject::GEODETIC, 252.224,
								  "NewNorcia"), 23, 453.17e-6);
													//	data provided by
													//	Ed Serpel on 2002/09/09

		_groundStations.push_back(Vilspa1);
		_groundStations.push_back(Vilspa2);
		_groundStations.push_back(Vilspa3);
		_groundStations.push_back(Perth);
		_groundStations.push_back(Kourou);
		_groundStations.push_back(Santiago);
		_groundStations.push_back(NewNorcia);
	}
} // XmmMission::checkGSset();


const XmmGroundStation &XmmMission::groundStation(
										XmmMission::GroundStationPlaces place)
{
	checkGSset();
	return _groundStations[UNT(place)];
} // XmmMission::groundStation()


class GSeq : public unary_function<XmmGroundStation, bool> {
	unsigned short _id;
public:
	explicit GSeq(unsigned short id) : _id(id) {}
	bool operator()(const XmmGroundStation &gs) { return gs.id()==_id; }
};


const XmmGroundStation &XmmMission::groundStation(unsigned short GSID)
{
	checkGSset();
	GroundStations::iterator i =
		find_if(_groundStations.begin(), _groundStations.end(), GSeq(GSID));
	if (i == _groundStations.end())
		errstr<<"There is no XMM ground station with ID "<<GSID
			  <<Sas::error("NoSuchGroundStation");

	return *i;
} // XmmMission::groundStation()
